from __future__ import annotations

from urllib.parse import urlparse, urlunparse

import smart_open
from airflow.providers.microsoft.azure.hooks.wasb import WasbHook
from azure.core.exceptions import ResourceNotFoundError

from astro.constants import FileLocation
from astro.files.locations.base import BaseFileLocation


class WASBLocation(BaseFileLocation):
    """Handler WASB object store operations"""

    location_type = FileLocation.WASB
    supported_conn_type = {WasbHook.conn_type, "wasbs"}

    def exists(self) -> bool:
        """Check if the file exists or not"""
        try:
            with smart_open.open(self.smartopen_uri, mode="r", transport_params=self.transport_params):
                return True
        except ResourceNotFoundError:
            return False

    @property
    def hook(self) -> WasbHook:
        return WasbHook(wasb_conn_id=self.conn_id) if self.conn_id else WasbHook()

    @property
    def transport_params(self) -> dict:
        """get WASB credentials for storage"""
        client = self.hook.get_conn()
        return {"client": client}

    @property
    def paths(self) -> list[str]:
        """Resolve WASB file paths with prefix"""
        url = urlparse(self.path)
        container_name = url.netloc
        prefix = url.path[1:]
        prefixes = self.hook.get_blobs_list(container_name=container_name, prefix=prefix)
        paths = [urlunparse((url.scheme, url.netloc, keys, "", "", "")) for keys in prefixes]
        return paths

    @property
    def smartopen_uri(self) -> str:
        """
        SmartOpen does not support URIs prefixed with wasb, so we need to change them to azure.

        :return: URI compatible with SmartOpen for Azure BlobStorage.
        """
        parsed_url = urlparse(self.path)
        if parsed_url.scheme == "wasbs":
            return self.path.replace("wasbs", "azure")
        elif parsed_url.scheme == "wasb":
            return self.path.replace("wasb", "azure")
        else:
            return self.path

    @property
    def size(self) -> int:
        """Return file size for WASB location"""
        url = urlparse(self.path)
        container_name = url.netloc
        object_name = url.path
        if object_name.startswith("/"):
            object_name = object_name[1:]
        return int(
            self.hook._get_blob_client(  # skipcq: PYL-W0212
                container_name=container_name, blob_name=object_name
            )
            .get_blob_properties()
            .size
        )

    @property
    def openlineage_dataset_namespace(self) -> str:
        """
        Returns the open lineage dataset namespace as per
        https://github.com/OpenLineage/OpenLineage/blob/main/spec/Naming.md
        """
        parsed_url = urlparse(self.path)
        return f"{parsed_url.scheme}://{parsed_url.netloc}"

    @property
    def openlineage_dataset_name(self) -> str:
        """
        Returns the open lineage dataset name as per
        https://github.com/OpenLineage/OpenLineage/blob/main/spec/Naming.md
        """
        return urlparse(self.path).path



from __future__ import annotations

from urllib.parse import urlparse

import smart_open
from airflow.providers.ftp.hooks.ftp import FTPHook

from astro.constants import FileLocation
from astro.files.locations.base import BaseFileLocation


class FTPLocation(BaseFileLocation):
    """Handler FTP object store operations"""

    location_type = FileLocation.FTP
    supported_conn_type = {FTPHook.conn_type}

    @property
    def hook(self) -> FTPHook:
        return FTPHook(ftp_conn_id=self.conn_id) if self.conn_id else FTPHook()

    @property
    def transport_params(self) -> dict:
        """get FTP credentials for remote file system"""
        return {}

    @property
    def paths(self) -> list[str]:
        """Resolve FTP file paths with prefix"""
        url = urlparse(self.path)
        conn = self.hook.get_connection(self.conn_id)
        uri = conn.get_uri()
        client = self.hook.get_conn()
        files = client.nlst("/" + url.netloc + url.path)
        # the above command checks if the path is a file or directory
        if len(files) > 0:
            paths = [uri + file for file in files]
        else:
            paths = [uri + "/" + url.netloc + url.path]
        return paths

    @property
    def size(self) -> int:
        """Return file size for FTP location"""
        url = urlparse(self.path)
        stat = self.hook.get_size(url.path)
        return int(stat) or -1

    @property
    def openlineage_dataset_namespace(self) -> str:
        """
        Returns the open lineage dataset namespace as per
        https://github.com/OpenLineage/OpenLineage/blob/main/spec/Naming.md
        """
        parsed_url = urlparse(self.path)
        return f"{parsed_url.scheme}://{parsed_url.netloc}"

    @property
    def openlineage_dataset_name(self) -> str:
        """
        Returns the open lineage dataset name as per
        https://github.com/OpenLineage/OpenLineage/blob/main/spec/Naming.md
        """
        return urlparse(self.path).path

    def get_uri(self):
        """Gets the URI of the connection"""
        conn = self.hook.get_connection(self.conn_id)
        return conn.get_uri()

    def get_stream(self):
        parsed_url = urlparse(self.path)
        path = f"{self.get_uri()}/{parsed_url.netloc}{parsed_url.path}"
        return smart_open.open(path, mode="wb", transport_params=self.transport_params)



from __future__ import annotations

import importlib
from pathlib import Path
from typing import TYPE_CHECKING

from astro.options import LoadOptionsList
from astro.utils.compat.functools import cache
from astro.utils.path import get_class_name, get_dict_with_module_names_to_dot_notations

if TYPE_CHECKING:  # pragma: no cover
    from astro.databases.base import BaseDatabase
    from astro.table import BaseTable

DEFAULT_CONN_TYPE_TO_MODULE_PATH = get_dict_with_module_names_to_dot_notations(Path(__file__))
CUSTOM_CONN_TYPE_TO_MODULE_PATH = {
    "gcpbigquery": DEFAULT_CONN_TYPE_TO_MODULE_PATH["bigquery"],
    "google_cloud_platform": DEFAULT_CONN_TYPE_TO_MODULE_PATH["bigquery"],
    "databricks": "astro.databases.databricks.delta",
    "delta": "astro.databases.databricks.delta",
}
CONN_TYPE_TO_MODULE_PATH = {
    **DEFAULT_CONN_TYPE_TO_MODULE_PATH,
    **CUSTOM_CONN_TYPE_TO_MODULE_PATH,
}
SUPPORTED_DATABASES = set(DEFAULT_CONN_TYPE_TO_MODULE_PATH.keys())


def create_database(
    conn_id: str, table: BaseTable | None = None, load_options_list: LoadOptionsList | None = None
) -> BaseDatabase:
    """
    Given a conn_id, return the associated Database class.

    :param conn_id: Database connection ID in Airflow
    :param table: (optional) The Table object
    """
    module = importlib.import_module(_get_conn(conn_id))
    class_name = get_class_name(module_ref=module, suffix="Database")
    database_class = getattr(module, class_name)
    load_options = load_options_list and load_options_list.get(database_class)
    database: BaseDatabase = database_class(conn_id, table, load_options=load_options)
    return database


@cache
def _get_conn(conn_id: str) -> str:
    from airflow.hooks.base import BaseHook

    conn_type = BaseHook.get_connection(conn_id).conn_type
    module_path = CONN_TYPE_TO_MODULE_PATH[conn_type]
    return module_path



from __future__ import annotations

import io

import pandas as pd

from astro.constants import FileType as FileTypeConstants
from astro.dataframes.load_options import PandasLoadOptions
from astro.dataframes.pandas import PandasDataframe
from astro.files.types.base import FileType
from astro.utils.dataframe import convert_columns_names_capitalization


class ParquetFileType(FileType):
    """Concrete implementation to handle Parquet file type"""

    LOAD_OPTIONS_CLASS_NAME = "PandasParquetLoadOptions"

    def export_to_dataframe(
        self,
        stream,
        columns_names_capitalization="original",
        **kwargs,
    ):
        """read parquet file from one of the supported locations and return dataframe

        :param stream: file stream object
        :param load_options: Pandas option to pass to the Pandas lib while reading parquet
        :param columns_names_capitalization: determines whether to convert all columns to lowercase/uppercase
            in the resulting dataframe
        """
        kwargs_copy = dict(kwargs)
        # Pandas `read_parquet` does not support the `nrows` parameter
        kwargs_copy.pop("nrows", None)

        byte_io_buffer = self._convert_remote_file_to_byte_stream(stream)
        if isinstance(self.load_options, PandasLoadOptions):
            kwargs_copy.update(self.load_options.to_dict())

        df = pd.read_parquet(byte_io_buffer, **kwargs_copy)
        df = convert_columns_names_capitalization(
            df=df, columns_names_capitalization=columns_names_capitalization
        )
        return PandasDataframe.from_pandas_df(df)

    @staticmethod
    def _convert_remote_file_to_byte_stream(stream) -> io.IOBase:
        """
        Convert file stream into a buffer that can be streamed into other data
        structures.
        Due to noted issues with using parquet files with smart_open+pandas (like
        https://github.com/RaRe-Technologies/smart_open/issues/524), we create a BytesIO buffer
        before exporting to a dataframe. We've found a sizable speed improvement with this optimization
        Returns: an io object that can be streamed into a dataframe (or other object)
        """
        remote_obj_buffer = io.BytesIO()
        remote_obj_buffer.write(stream.read())
        remote_obj_buffer.seek(0)
        return remote_obj_buffer

    # We need skipcq because it's a method overloading so we don't want to make it a static method
    def create_from_dataframe(self, df: pd.DataFrame, stream: io.TextIOWrapper) -> None:  # skipcq PYL-R0201
        """Write parquet file to one of the supported locations

        :param df: pandas dataframe
        :param stream: file stream object
        """
        df.to_parquet(stream)

    @property
    def name(self):
        return FileTypeConstants.PARQUET



from __future__ import annotations

import pathlib

from astro.constants import FileType as FileTypeConstants
from astro.files.types.base import FileType
from astro.files.types.csv import CSVFileType
from astro.files.types.json import JSONFileType
from astro.files.types.ndjson import NDJSONFileType
from astro.files.types.parquet import ParquetFileType
from astro.options import LoadOptionsList


def create_file_type(
    path: str,
    filetype: FileTypeConstants | None = None,
    normalize_config: dict | None = None,
    load_options_list: LoadOptionsList | None = None,
) -> FileType:
    """Factory method to create FileType super objects based on the file extension in path or filetype specified."""
    filetype_to_class: dict[FileTypeConstants, type[FileType]] = {
        FileTypeConstants.CSV: CSVFileType,
        FileTypeConstants.JSON: JSONFileType,
        FileTypeConstants.NDJSON: NDJSONFileType,
        FileTypeConstants.PARQUET: ParquetFileType,
    }
    if not filetype:
        filetype = get_filetype(path)

    try:
        load_options = None
        if load_options_list is not None:
            load_options = load_options_list.get(filetype_to_class[filetype])

        return filetype_to_class[filetype](
            path=path,
            normalize_config=normalize_config,
            load_options=load_options,
        )
    except KeyError:
        raise ValueError(
            f"Non supported file type provided {filetype}, file_type should be among {', '.join(FileTypeConstants)}."
        )


def get_filetype(filepath: str | pathlib.PosixPath) -> FileTypeConstants:
    """
    Return a FileType given the filepath. Uses a naive strategy, using the file extension.

    :param filepath: URI or Path to a file
    :type filepath: str or pathlib.PosixPath
    :return: The filetype (e.g. csv, ndjson, json, parquet)
    :rtype: astro.constants.FileType
    """
    if isinstance(filepath, pathlib.PosixPath):
        extension = filepath.suffix[1:]
    else:
        extension = ""
        tokenized_path = filepath.split(".")
        if len(tokenized_path) > 1:
            extension = tokenized_path[-1]

    if extension == "":
        raise ValueError(
            f"Missing file extension, cannot automatically determine filetype from path '{filepath}'."
            f" Please pass the 'filetype' param with the explicit filetype (e.g. csv, ndjson, etc.)."
        )

    try:
        return FileTypeConstants(extension)
    except ValueError:
        raise ValueError(f"Unsupported filetype '{extension}' from file '{filepath}'.")



from typing import Any, Dict, Optional

from airflow.decorators.base import get_unique_task_id
from airflow.models.xcom_arg import XComArg
from airflow.providers.common.sql.hooks.sql import DbApiHook
from airflow.providers.common.sql.operators.sql import SQLTableCheckOperator

from astro.databases import create_database
from astro.table import BaseTable
from astro.utils.compat.typing import Context


class SQLCheckOperator(SQLTableCheckOperator):
    """
    Performs one or more of the checks provided in the checks dictionary.
    Checks should be written to return a boolean result.

    :param dataset: the table to run checks on
    :param checks: the dictionary of checks, e.g.:

    .. code-block:: python

        {
            "row_count_check": {"check_statement": "COUNT(*) = 1000"},
            "column_sum_check": {"check_statement": "col_a + col_b < col_c"},
        }


    :param partition_clause: a partial SQL statement that is added to a WHERE clause in the query built by
        the operator that creates partition_clauses for the checks to run on, e.g.

    .. code-block:: python

        "date = '1970-01-01'"
    """

    template_fields = ("partition_clause", "dataset")

    def __init__(
        self,
        *,
        dataset: BaseTable,
        checks: Dict[str, Dict[str, Any]],
        partition_clause: Optional[str] = None,
        task_id: Optional[str] = None,
        **kwargs,
    ):

        self.dataset = dataset
        super().__init__(
            table="place_holder_table_name",
            checks=checks,
            partition_clause=partition_clause,
            task_id=task_id or get_unique_task_id("check_table"),
        )

    def execute(self, context: "Context"):
        db = create_database(self.dataset.conn_id)
        self.table = db.get_table_qualified_name(self.dataset)
        self.conn_id = self.dataset.conn_id
        # apache-airflow-providers-common-sql == 1.2.0 which is compatible with airflow 2.2.5 implements the self.sql
        # differently compared to apache-airflow-providers-common-sql == 1.3.3
        try:
            self.sql = f"SELECT check_name, check_result FROM ({self._generate_sql_query()}) AS check_table"
        except AttributeError:
            self.sql = f"SELECT * FROM {self.table};"
        super().execute(context)

    def get_db_hook(self) -> DbApiHook:
        """
        Get the database hook for the connection.

        :return: the database hook object.
        """
        db = create_database(conn_id=self.conn_id)
        if db.sql_type == "bigquery":
            return db.hook
        return super().get_db_hook()


def check_table(
    dataset: BaseTable,
    checks: Dict[str, Dict[str, Any]],
    partition_clause: Optional[str] = None,
    task_id: Optional[str] = None,
    **kwargs,
) -> XComArg:
    """
    Performs one or more of the checks provided in the checks dictionary.
    Checks should be written to return a boolean result.

    :param dataset: the table to run checks on
    :param checks: the dictionary of checks, e.g.:

    .. code-block:: python

        {
            "row_count_check": {"check_statement": "COUNT(*) = 1000"},
            "column_sum_check": {"check_statement": "col_a + col_b < col_c"},
        }


    :param partition_clause: a partial SQL statement that is added to a WHERE clause in the query built by
        the operator that creates partition_clauses for the checks to run on, e.g.

    .. code-block:: python

        "date = '1970-01-01'"
    """
    return SQLCheckOperator(
        dataset=dataset,
        checks=checks,
        partition_clause=partition_clause,
        kwargs=kwargs,
        task_id=task_id,
    ).output



from __future__ import annotations

from collections.abc import Iterable, Mapping
from typing import Any, Callable

try:
    from airflow.decorators.base import TaskDecorator
except ImportError:
    from airflow.decorators import _TaskDecorator as TaskDecorator

from airflow.decorators.base import get_unique_task_id, task_decorator_factory
from airflow.models.xcom_arg import XComArg
from sqlalchemy.sql.functions import Function

from astro.sql.operators.base_decorator import BaseSQLDecoratedOperator
from astro.utils.compat.typing import Context


class TransformOperator(BaseSQLDecoratedOperator):
    """
    Given a SQL statement and (optional) tables, execute the SQL statement and output
    the result into a SQL table.
    """

    def __init__(
        self,
        conn_id: str | None = None,
        parameters: dict | None = None,
        handler: Function | None = None,
        database: str | None = None,
        schema: str | None = None,
        response_limit: int = -1,
        response_size: int = -1,
        sql: str = "",
        task_id: str = "",
        **kwargs: Any,
    ):
        task_id = task_id or get_unique_task_id("transform")
        super().__init__(
            conn_id=conn_id,
            parameters=parameters,
            handler=handler,
            database=database,
            schema=schema,
            response_limit=response_limit,
            response_size=response_size,
            sql=sql,
            task_id=task_id,
            **kwargs,
        )

    def execute(self, context: Context):
        super().execute(context)
        self.database_impl.create_schema_if_needed(self.output_table.metadata.schema)
        self.database_impl.drop_table(self.output_table)
        self.database_impl.create_table_from_select_statement(
            statement=self.sql,
            target_table=self.output_table,
            parameters=self.parameters,
        )
        # TODO: remove pushing to XCom once we update the airflow version.
        context["ti"].xcom_push(
            key="output_table_row_count", value=str(self.database_impl.row_count(self.output_table))
        )
        context["ti"].xcom_push(key="output_table_conn_id", value=str(self.output_table.conn_id))
        return self.output_table


def transform(
    python_callable: Callable | None = None,
    conn_id: str = "",
    parameters: Mapping | Iterable | None = None,
    database: str | None = None,
    schema: str | None = None,
    **kwargs: Any,
) -> TaskDecorator:
    """
    Given a python function that returns a SQL statement and (optional) tables, execute the SQL statement and output
    the result into a SQL table.

    Use this function as a decorator like so:

    .. code-block:: python

      @transform
      def my_sql_statement(table1: Table, table2: Table) -> Table:
          return "SELECT * FROM {{table1}} JOIN {{table2}}"

    In this example, by identifying the parameters as `Table` objects, astro knows to automatically convert those
    objects into tables (if they are, for example, a dataframe). Any type besides table will lead astro to assume
    you do not want the parameter converted.

    You can also pass parameters into the query like so

     .. code-block:: python

      @transform
      def my_sql_statement(table1: Table, table2: Table, execution_date) -> Table:
          return "SELECT * FROM {{table1}} JOIN {{table2}} WHERE date > {{exec_date}}", {
              "exec_date": execution_date
          }

    :param python_callable: This parameter is filled in automatically when you use the transform function as a decorator
        This is where the python function gets passed to the wrapping function
    :param conn_id: Connection ID for the database you want to connect to. If you do not pass in a value for this object
        we can infer the connection ID from the first table passed into the python_callable function.
        (required if there are no table arguments)
    :param parameters: parameters to pass into the SQL query
    :param database: Database within the SQL instance you want to access. If left blank we will default to the
        table.metatadata.database in the first Table passed to the function (required if there are no table arguments)
    :param schema: Schema within the SQL instance you want to access. If left blank we will default to the
        table.metatadata.schema in the first Table passed to the function (required if there are no table arguments)
    :param kwargs: Any keyword arguments supported by the BaseOperator is supported (e.g ``queue``, ``owner``)
    :return: Transform functions return a ``Table`` object that can be passed to future tasks.
        This table will be either an auto-generated temporary table,
        or will overwrite a table given in the `output_table` parameter.
    """

    kwargs.update(
        {
            "conn_id": conn_id,
            "parameters": parameters,
            "database": database,
            "schema": schema,
            "handler": None,
        }
    )
    return task_decorator_factory(
        python_callable=python_callable,
        multiple_outputs=False,
        decorated_operator_class=TransformOperator,
        **kwargs,
    )


def transform_file(
    file_path: str,
    conn_id: str = "",
    parameters: dict | None = None,
    database: str | None = None,
    schema: str | None = None,
    **kwargs: Any,
) -> XComArg:
    """
    This function returns a ``Table`` object that can be passed to future tasks from specified SQL file.
    Tables can be inserted via the parameters kwarg.

    :param file_path: File path for the SQL file you would like to parse. Can be an absolute path, or you can use a
        relative path if the `template_searchpath` variable is set in your DAG
    :param conn_id: Connection ID for the database you want to connect to. If you do not pass in a value for this object
        we can infer the connection ID from the first table passed into the python_callable function.
        (required if there are no table arguments)
    :param parameters: parameters to pass into the SQL query
    :param database: Database within the SQL instance you want to access. If left blank we will default to the
        table.metatadata.database in the first Table passed to the function (required if there are no table arguments)
    :param schema: Schema within the SQL instance you want to access. If left blank we will default to the
        table.metatadata.schema in the first Table passed to the function (required if there are no table arguments)
    :param kwargs: Any keyword arguments supported by the BaseOperator is supported (e.g ``queue``, ``owner``)
    :return: Transform functions return a ``Table`` object that can be passed to future tasks.
        This table will be either an auto-generated temporary table,
        or will overwrite a table given in the `output_table` parameter.
    """

    kwargs.update(
        {
            "op_kwargs": kwargs.get("op_kwargs", {}),
            "op_args": kwargs.get("op_args", {}),
        }
    )
    return TransformOperator(
        conn_id=conn_id,
        parameters=parameters,
        handler=None,
        database=database,
        schema=schema,
        sql=file_path,
        python_callable=lambda: (file_path, parameters),
        **kwargs,
    ).output



from __future__ import annotations

import glob
import os
from abc import ABC, abstractmethod
from pathlib import Path
from urllib.parse import urlparse

import smart_open
from airflow.hooks.base import BaseHook

from astro.constants import FileLocation
from astro.options import LoadOptions


class BaseFileLocation(ABC):
    """Base Location abstract class"""

    template_fields = ("path", "conn_id")
    supported_conn_type: set[str] = set()

    def __init__(self, path: str, conn_id: str | None = None, load_options: LoadOptions | None = None):
        """
        Manages and provide interface for the operation for all the supported locations.

        :param path: Path to a file in the filesystem/Object stores
        :param conn_id: Airflow connection ID
        """
        self.path: str = path
        self.conn_id: str | None = conn_id
        self.load_options: LoadOptions | None = load_options
        self.validate_conn()

    def validate_conn(self):
        """Check if the conn_id matches with provided path."""
        if not self.conn_id:
            return

        connection_type = BaseHook.get_connection(self.conn_id).conn_type
        if connection_type not in self.supported_conn_type:
            raise ValueError(
                f"Connection type {connection_type} is not supported for {self.path}. "
                f"Supported types are {self.supported_conn_type}"
            )

    @property
    def smartopen_uri(self) -> str:
        """
        Changes the object URI (self.path) to a SmartOpen supported URI if necessary.
        By default, does not change the self.path.

        :return: URI compatible with SmartOpen for desired location.
        """
        return self.path

    @property
    def hook(self):
        raise NotImplementedError

    @property
    @abstractmethod
    def location_type(self) -> FileLocation:
        """Property to identify location type"""
        raise NotImplementedError

    @property
    @abstractmethod
    def paths(self) -> list[str]:
        """Resolve patterns in path"""
        raise NotImplementedError

    @property
    def transport_params(self) -> dict | None:  # skipcq: PYL-R0201
        """Get credentials required by smart open to access files"""
        return None

    @property
    @abstractmethod
    def size(self) -> int:
        """Return the size in bytes of the given file"""
        raise NotImplementedError

    @property
    @abstractmethod
    def openlineage_dataset_namespace(self) -> str:
        """
        Returns the open lineage dataset namespace as per
        https://github.com/OpenLineage/OpenLineage/blob/main/spec/Naming.md
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def openlineage_dataset_name(self) -> str:
        """
        Returns the open lineage dataset name as per
        https://github.com/OpenLineage/OpenLineage/blob/main/spec/Naming.md
        """
        raise NotImplementedError

    @staticmethod
    def is_valid_path(path: str) -> bool:
        """
        Check if the given path is either a valid URI or a local file

        :param path: Either local filesystem path or remote URI
        """
        try:
            BaseFileLocation.get_location_type(path)
        except ValueError:
            return False

        try:
            result = urlparse(path)

            if not (
                (result.scheme and result.netloc and (result.port or result.port is None))
                or os.path.isfile(path)
                or BaseFileLocation.check_non_existing_local_file_path(path)
                or glob.glob(result.path)
            ):
                return False

            return True
        except ValueError:
            return False

    @staticmethod
    def check_non_existing_local_file_path(path: str) -> bool:
        """Check if the path is valid by creating and temp file and then deleting it. Assumes the file don't exist"""
        try:
            Path(path).touch()
            os.remove(path)
        except OSError:
            return False
        return True

    @staticmethod
    def get_location_type(path: str) -> FileLocation:
        """Identify where a file is located

        :param path: Path to a file in the filesystem/Object stores
        """
        file_scheme = urlparse(path).scheme
        if file_scheme == "":
            location = FileLocation.LOCAL
        else:
            try:
                location = FileLocation(file_scheme)
            except ValueError:
                raise ValueError(f"Unsupported scheme '{file_scheme}' from path '{path}'")
        return location

    def exists(self) -> bool:
        """Check if the file exists or not"""
        try:
            with smart_open.open(self.smartopen_uri, mode="r", transport_params=self.transport_params):
                return True
        except OSError:
            return False

    def databricks_auth_settings(self) -> dict:
        """
        Required settings to upload this file into databricks. Only needed for cloud storage systems
        like S3
        :return: A dictionary of settings keys to settings values
        """
        return {}

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(path="{self.path}",conn_id="{self.conn_id}")'

    def __str__(self) -> str:
        """String representation of location"""
        return self.path

    def __eq__(self, other) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.path == other.path and self.conn_id == other.conn_id

    def __hash__(self) -> int:
        return hash((self.path, self.conn_id))

    def get_stream(self):
        """Create a file in the desired location using the smart_open.

        :param df: pandas dataframe
        """
        return smart_open.open(self.smartopen_uri, mode="wb", transport_params=self.transport_params)



from __future__ import annotations

import logging
import pathlib
import tempfile
from pathlib import Path
from typing import Any

from airflow.configuration import conf
from databricks_cli.sdk.api_client import ApiClient

from astro.constants import DatabricksLoadMode, FileLocation, LoadExistStrategy
from astro.databases.databricks.api_utils import (
    create_and_run_job,
    create_secrets,
    delete_secret_scope,
    generate_file,
    load_file_to_dbfs,
)
from astro.databases.databricks.load_options import DeltaLoadOptions
from astro.files import File
from astro.table import BaseTable

cwd = pathlib.Path(__file__).parent

supported_file_locations = [FileLocation.LOCAL, FileLocation.S3, FileLocation.GS]

log = logging.getLogger(__file__)


def load_file_to_delta(  # noqa: C901
    input_file: File,
    delta_table: BaseTable,
    databricks_job_name: str,
    delta_load_options: DeltaLoadOptions,
    if_exists: LoadExistStrategy = "replace",
):
    """
    Load a file object into a databricks delta table

    :param if_exists: How to handle the data if the table exists. Defaults to replace (but can be set to "append")
    :param delta_load_options: Loading options for passing data into delta. Things like COPY_OPTIONS, cluster_id, etc.
    :param databricks_job_name: The name of the job as it will show up in databricks.
    :param input_file: File to load into delta
    :param delta_table: a Table object with necessary metadata for accessing the cluster.
    """
    from astro.databases.databricks.delta import DeltaDatabase

    delta_load_options.if_exists = if_exists
    db = DeltaDatabase(conn_id=delta_table.conn_id)
    api_client = db.api_client
    if input_file.location.location_type not in supported_file_locations:
        raise ValueError(
            f"File location {input_file.location.location_type} not yet supported. "
            f"Currently supported locations: {','.join([str(s) for s in supported_file_locations])}"
        )
    dbfs_file_path = None
    if input_file.is_local():
        # There are no secrets to load since this is a local file
        log.info("Found local file. Uploading file(s) to DBFS")
        delta_load_options.load_secrets = False
        dbfs_file_path = _load_local_file_to_dbfs(api_client, input_file)
    elif delta_load_options.load_secrets and conf.getboolean(
        "astro_sdk", "load_storage_configs_to_databricks", fallback=True
    ):
        log.info("Loading storage configurations for remote file to databricks")
        delete_secret_scope(delta_load_options.secret_scope, api_client=api_client)
        _load_secrets_to_databricks(api_client, input_file, delta_load_options.secret_scope)

    if not input_file.is_directory() and delta_load_options.load_mode == DatabricksLoadMode.AUTOLOADER:
        log.info("Since this is a single file, changing to COPY INTO as autoloader only supports directories")
        delta_load_options.load_mode = DatabricksLoadMode.COPY_INTO

    with tempfile.NamedTemporaryFile(suffix=".py") as tfile:
        delta_load_options.if_exists = if_exists

        log.info("Generating Pyspark file")
        dbfs_file_path = _create_load_file_pyspark_file(
            api_client, delta_load_options, dbfs_file_path, delta_table, input_file, tfile
        )
    try:
        create_and_run_job(
            api_client=api_client,
            databricks_job_name=databricks_job_name,
            existing_cluster_id=delta_load_options.existing_cluster_id,
            file_to_run=str(dbfs_file_path),
        )
    finally:
        if (
            input_file.location.location_type != FileLocation.LOCAL
            and delta_load_options.secret_scope == DeltaLoadOptions.get_default_delta_options().secret_scope
        ):
            delete_secret_scope(delta_load_options.secret_scope, api_client=api_client)


def _create_load_file_pyspark_file(
    api_client: ApiClient,
    databricks_options: DeltaLoadOptions,
    dbfs_file_path: Path | None,
    delta_table: BaseTable,
    input_file: File,
    output_file: Any,
):
    file_type = _find_file_type(input_file)
    if (
        databricks_options.load_mode == DatabricksLoadMode.AUTOLOADER
        and databricks_options.if_exists == "replace"
    ):
        databricks_options.autoloader_load_options["cloudFiles.allowOverwrites"] = "true"

    file_path = generate_file(
        data_source_path=str(dbfs_file_path) if dbfs_file_path else input_file.path,
        table_name=delta_table.name,
        source_type=str(input_file.location.location_type),
        output_file_path=Path(output_file.name),
        file_type=file_type or "",
        load_options=databricks_options,
    )
    dbfs_file_path = load_file_to_dbfs(
        file_path, file_name=f"load_file_{delta_table.name}.py", api_client=api_client
    )
    return dbfs_file_path


def _load_local_file_to_dbfs(api_client: ApiClient, input_file: File):
    file_path = Path(input_file.path)
    dbfs_file_path = load_file_to_dbfs(
        local_file_path=Path(input_file.path), file_name=file_path.name, api_client=api_client
    )
    return dbfs_file_path


def _load_secrets_to_databricks(api_client: ApiClient, input_file: File, secret_scope_name: str):
    """
    Load secrets into a secret scope in databricks. This can be used as a utility for loading
    necessary secrets for accessing external data stores.

    :param api_client: Databricks API client loaded with credentials
    :param input_file: file for loading connection info
    :param secret_scope_name: name of the secret scope to load info into
    """
    create_secrets(
        scope_name=secret_scope_name,
        filesystem_secrets=input_file.location.databricks_auth_settings(),
        api_client=api_client,
    )


def _find_file_type(input_file: File) -> str:
    """
    Since COPY INTO does not automatically detect schema, we need to ensure that the file type is
    either set by the user, or inferable in the URL
    :param input_file: file to parse
    :return: file type of the uploaded file
    """
    if input_file.filetype:
        return str(input_file.filetype)
    if not input_file.filetype and "." in input_file.path:
        return input_file.path.split(".")[-1]
    raise ValueError("For COPY INTO, you need to supply a file type.")



from __future__ import annotations

from typing import Any

from airflow.decorators.base import get_unique_task_id
from airflow.models import BaseOperator

from astro.files.base import File
from astro.files.locations import create_file_location
from astro.utils.compat.typing import Context


class ListFileOperator(BaseOperator):
    """
    List the file available at path and storage

    :param task_id: The task id for uniquely identify a task in a DAG
    :param path: A path pattern for which you want to get a list of file
    :param conn_id: Airflow connection id.
        This will be used to identify the right Airflow hook at runtime to connect with storage services
    """

    template_fields = ("path", "conn_id")

    def __init__(self, path: str, conn_id: str, task_id: str = "", **kwargs):
        task_id = task_id or get_unique_task_id(
            "get_file_list", dag=kwargs.get("dag"), task_group=kwargs.get("task_group")
        )
        self.path = path
        self.conn_id = conn_id
        super().__init__(task_id=task_id, **kwargs)

    def execute(self, context: Context) -> Any:  # skipcq: PYL-W0613
        files = get_file_list_func(self.path, self.conn_id)
        # Get list of files excluding folders
        return [File(path=file, conn_id=self.conn_id) for file in files]


def get_file_list_func(path: str, conn_id: str) -> list[str]:
    """Function to get list of files from a bucket"""
    location = create_file_location(path, conn_id)
    # Get list of files excluding folders
    return [path for path in location.paths if not path.endswith("/")]



from typing import Dict, List, Optional

import attr


@attr.define
class LoadOptions:
    def empty(self):
        return NotImplementedError()

    def to_dict(self) -> dict:
        """
        Convert options class to dict
        """
        return attr.asdict(self)


def list_to_dict(value: Optional[List[LoadOptions]]) -> Optional[Dict[str, LoadOptions]]:
    """
    Convert list object to dict
    """
    if value is None:
        return None
    return {type(option).__name__: option for option in value}


@attr.define
class LoadOptionsList:
    _load_options: Optional[Dict[str, LoadOptions]] = attr.field(converter=list_to_dict)

    def get(self, option_class) -> Optional[LoadOptions]:
        """
        Check `LOAD_OPTIONS_CLASS_NAME` attribute and select the correct load_options
        :param option_class: FileType | FileLocation | BaseDatabase
        """
        if not hasattr(option_class, "LOAD_OPTIONS_CLASS_NAME"):
            return None
        return self.get_by_class_name(option_class.LOAD_OPTIONS_CLASS_NAME)

    def get_by_class_name(self, option_class_name) -> Optional[LoadOptions]:
        """
        Get load_option by class name
        :return:
        """
        if self._load_options is None:
            return None
        return self._load_options.get(option_class_name, None)


@attr.define
class SnowflakeLoadOptions(LoadOptions):
    """
    Load options to load file to snowflake using native approach.

    :param file_options: Depending on the file format type specified, use one or more of the
        format-specific options as key-value pair. Read more at:
        https://docs.snowflake.com/en/sql-reference/sql/copy-into-table.html#format-type-options-formattypeoptions
    :param copy_options: Specify one or more of the copy option as key-value pair. Read more at:
        https://docs.snowflake.com/en/sql-reference/sql/copy-into-table.html#copy-options-copyoptions
    """

    file_options: dict = attr.field(init=True, factory=dict)
    copy_options: dict = attr.field(init=True, factory=dict)
    storage_integration: str = attr.field(default=None)

    def empty(self):
        return not self.file_options and not self.copy_options



from __future__ import annotations

import io
from abc import ABC, abstractmethod

import pandas as pd

from astro.options import LoadOptions


class FileType(ABC):
    """Abstract File type class, meant to be the interface to all client code for all supported file types"""

    def __init__(
        self, path: str, normalize_config: dict | None = None, load_options: LoadOptions | None = None
    ):
        self.path = path
        self.normalize_config = normalize_config
        self.load_options = load_options

    @abstractmethod
    def export_to_dataframe(self, stream, **kwargs) -> pd.DataFrame:
        """read file from one of the supported locations and return dataframe

        :param stream: file stream object
        """
        raise NotImplementedError

    @abstractmethod
    def create_from_dataframe(self, df: pd.DataFrame, stream: io.TextIOWrapper) -> None:
        """Write file to one of the supported locations

        :param df: pandas dataframe
        :param stream: file stream object
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def name(self):
        """get file type"""
        raise NotImplementedError

    def __str__(self):
        """String representation of type"""
        return self.name.value

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(path="{self.path}")'

    def __eq__(self, other):
        return self.name == other.name

    def __hash__(self):
        return hash(self.name)



from __future__ import annotations

import io
from contextlib import contextmanager
from urllib.parse import urlparse

from airflow.providers.sftp.hooks.sftp import SFTPHook

from astro.constants import FileLocation
from astro.exceptions import PermissionNotSetError
from astro.files.locations.base import BaseFileLocation


class SFTPLocation(BaseFileLocation):
    """
    Handler SFTP object store operations.
    While exporting to SFTP the user should have permission
    to create the file in the desired location.
    """

    location_type = FileLocation.SFTP
    supported_conn_type = {SFTPHook.conn_type}

    @property
    def hook(self) -> SFTPHook:
        return SFTPHook(ssh_conn_id=self.conn_id) if self.conn_id else SFTPHook()

    @property
    def transport_params(self) -> dict:
        """Get SFTP credentials in order to access remote file system"""
        client = self.hook.get_connection(self.conn_id)
        extra_options = client.extra_dejson
        if "key_file" in extra_options:
            key_file = extra_options.get("key_file")
            return {"connect_kwargs": {"key_filename": key_file}}
        elif client.password:
            return {"connect_kwargs": {"password": client.password}}
        raise PermissionNotSetError("SFTP credentials are not set in the connection.")

    @property
    def paths(self) -> list[str]:
        """Resolve SFTP file paths with prefix"""
        url = urlparse(self.path)
        uri = self.get_uri()
        full_paths = []
        prefixes = self.hook.get_tree_map(url.netloc, prefix=url.netloc + url.path)
        for keys in prefixes:
            if len(keys) > 0:
                full_paths.extend(keys)
        paths = [uri + "/" + path for path in full_paths]
        return paths

    @property
    def size(self) -> int:
        """Return file size for SFTP location"""
        url = urlparse(self.path)
        conn = self.hook.get_conn()
        stat = conn.stat(url.netloc + url.path).st_size
        return int(stat) or -1

    @property
    def openlineage_dataset_namespace(self) -> str:
        """
        Returns the open lineage dataset namespace as per
        https://github.com/OpenLineage/OpenLineage/blob/main/spec/Naming.md
        """
        parsed_url = urlparse(self.path)
        return f"{parsed_url.scheme}://{parsed_url.netloc}"

    @property
    def openlineage_dataset_name(self) -> str:
        """
        Returns the open lineage dataset name as per
        https://github.com/OpenLineage/OpenLineage/blob/main/spec/Naming.md
        """
        return urlparse(self.path).path

    def get_uri(self):
        client = self.hook.get_connection(self.conn_id)
        return client.get_uri()

    def get_stream(self):
        """return the custom SFTP read_buffer context to add file into given path"""
        return self.sftp_stream()

    @contextmanager
    def sftp_stream(self):
        """sftp_stream context to export a file to the given location."""
        buffer = io.BytesIO()
        try:
            yield buffer
        finally:
            buffer.seek(0)
            sftp = self.hook.get_conn()
            parsed_url = urlparse(self.path)
            sftp.putfo(buffer, parsed_url.netloc + parsed_url.path)
            buffer.close()



from __future__ import annotations

import tempfile
import uuid
import warnings
from textwrap import dedent

import pandas as pd
from airflow.providers.databricks.hooks.databricks import DatabricksHook
from airflow.providers.databricks.hooks.databricks_sql import DatabricksSqlHook
from databricks.sql.client import Cursor
from databricks_cli.sdk.api_client import ApiClient
from sqlalchemy.engine.base import Engine as SqlAlchemyEngine
from sqlalchemy.sql import ClauseElement

from astro.constants import DEFAULT_CHUNK_SIZE, ColumnCapitalization, LoadExistStrategy, MergeConflictStrategy
from astro.databases.base import BaseDatabase
from astro.databases.databricks.load_file.load_file_job import load_file_to_delta
from astro.databases.databricks.load_options import DeltaLoadOptions
from astro.dataframes.pandas import PandasDataframe
from astro.files import File
from astro.options import LoadOptions
from astro.table import BaseTable, Metadata


class DeltaDatabase(BaseDatabase):

    LOAD_OPTIONS_CLASS_NAME = "DeltaLoadOptions"
    # In run_raw_sql operator decides if we want to return results directly or process them by handler provided
    # For delta tables we ignore the handler
    IGNORE_HANDLER_IN_RUN_RAW_SQL: bool = True
    _create_table_statement: str = "CREATE TABLE IF NOT EXISTS {} USING DELTA AS {} "

    def __init__(self, conn_id: str, table: BaseTable | None = None, load_options: LoadOptions | None = None):
        super().__init__(conn_id)
        self.table = table
        self.load_options = load_options or DeltaLoadOptions.get_default_delta_options()

    def populate_table_metadata(self, table: BaseTable) -> BaseTable:
        # TODO: Do we need default configurations for a delta table?
        """
        Given a table, populates the "metadata" field with what we would consider as "defaults"
        These defaults are determined based on environment variables and the connection settings.

        :param table: table to be populated
        :return:
        """
        table.conn_id = table.conn_id or self.conn_id
        if not table.metadata or table.metadata.is_empty():
            if self.table:
                table.metadata = self.table.metadata
            else:
                table.metadata = self.default_metadata
        return table

    @property
    def api_client(self) -> ApiClient:
        """
        Returns the databricks API client. Used for interacting with databricks services like
        DBFS, Jobs, etc.

        :return: A databricks ApiClient
        """
        conn = DatabricksHook(
            databricks_conn_id=self.conn_id
        ).get_conn()  # add this because DatabricksSqlHook does not expose password
        api_client = ApiClient(host=conn.host, token=conn.password)
        return api_client

    @property
    def sql_type(self):
        return "delta"

    @property
    def hook(self) -> DatabricksSqlHook:
        """
        Return the hook for the relevant databricks conn_id

        :return: a DatabricksSqlHook with metadata
        """
        return DatabricksSqlHook(databricks_conn_id=self.conn_id)

    @property
    def sqlalchemy_engine(self) -> SqlAlchemyEngine:
        raise NotImplementedError("We are not using sqlalchemy for databricks")

    @property
    def default_metadata(self) -> Metadata:
        return Metadata()

    def create_table_using_native_schema_autodetection(self, table: BaseTable, file: File) -> None:
        # TODO Do we need to implement this function? It seems like databricks will handle schemas for us
        raise NotImplementedError("Not implemented yet.")

    def schema_exists(self, schema: str) -> bool:
        # Schemas do not need to be created for delta, so we can assume this is true
        return True

    def create_schema_if_needed(self, schema: str | None) -> None:  # skipcq: PYL-W0613
        # Schemas do not need to be created for delta, so we don't need to do anything here
        return None

    def fetch_all_rows(self, table: BaseTable, row_limit: int = -1) -> list:
        """
        Fetches all rows for a table and returns as a list. This is needed because some
        databases have different cursors that require different methods to fetch rows

        :param table: The table metadata needed to fetch the rows
        :param row_limit: Limit the number of rows returned, by default return all rows.
        :return: a list of rows
        """
        statement = f"SELECT * FROM {self.get_table_qualified_name(table)};"
        if row_limit > -1:
            statement = statement + f" LIMIT {row_limit}"
        return self.run_sql(statement, handler=lambda x: x.fetchall())  # type: ignore

    def load_file_to_table(
        self,
        input_file: File,
        output_table: BaseTable,
        normalize_config: dict | None = None,
        if_exists: LoadExistStrategy = "replace",
        chunk_size: int = DEFAULT_CHUNK_SIZE,
        use_native_support: bool = True,
        native_support_kwargs: dict | None = None,
        columns_names_capitalization: ColumnCapitalization = "original",
        enable_native_fallback: bool | None = None,
        databricks_job_name: str = "",
        **kwargs,
    ):
        """
        Load content of multiple files in output_table.
        Multiple files are sourced from the file path, which can also be path pattern.

        :param databricks_job_name: Create a consistent job name so that we don't litter the databricks job screen.
            This should be <dag_id>_<task_id>
        :param input_file: File path and conn_id for object stores
        :param output_table: Table to create
        :param if_exists: Overwrite file if exists
        :param chunk_size: Specify the number of records in each batch to be written at a time
        :param use_native_support: Use native support for data transfer if available on the destination
        :param normalize_config: pandas json_normalize params config
        :param native_support_kwargs: kwargs to be used by method involved in native support flow
        :param columns_names_capitalization: determines whether to convert all columns to lowercase/uppercase
            in the resulting dataframe
        :param enable_native_fallback: Use enable_native_fallback=True to fall back to default transfer

        """
        load_file_to_delta(
            input_file=input_file,
            delta_table=output_table,
            databricks_job_name=databricks_job_name,
            delta_load_options=self.load_options,  # type: ignore
            if_exists=if_exists,
        )

    def openlineage_dataset_name(self, table: BaseTable) -> str:
        return ""

    def openlineage_dataset_namespace(self) -> str:
        return ""

    def create_table_from_select_statement(
        self,
        statement: str,
        target_table: BaseTable,
        parameters: dict | None = None,
    ) -> None:
        """
        Create a Delta table from a SQL SELECT statement.

        :param statement: Statement that will return a table
        :param target_table: The table which the result of the SQL statement will be placed
        :param parameters: Parameters to pass to databricks
        :return: None
        """
        statement = self._create_table_statement.format(
            self.get_table_qualified_name(target_table), statement
        )
        self.run_sql(sql=statement, parameters=parameters)

    def parameterize_variable(self, variable: str) -> str:
        """
        Parameterize a variable in a way that the Databricks SQL API can recognize.
        :param variable: Variable to parameterize
        :return: The number of rows in the table
        """
        return "%(" + variable + ")s"

    def row_count(self, table: BaseTable):
        result = self.run_sql(
            f"SELECT COUNT(*) FROM {self.get_table_qualified_name(table)}",  # skipcq: BAN-B608
            handler=lambda x: x.fetchone(),
        )
        return result.asDict()["count(1)"]

    def run_sql(
        self,
        sql: str | ClauseElement = "",
        parameters: dict | None = None,
        handler=None,
        **kwargs,
    ):
        """
        Run SQL against a delta table using spark SQL.

        :param sql: SQL Query to run on delta table
        :param parameters: parameters to pass to delta
        :param handler: function that takes in a databricks cursor as an argument.
        :param kwargs:
        :return: None if there is no handler, otherwise return result of handler function
        """
        hook = DatabricksSqlHook(
            databricks_conn_id=self.conn_id,
        )
        return hook.run(sql, parameters=parameters, handler=handler)

    def table_exists(self, table: BaseTable) -> bool:
        """
        Queries databricks to check if a table exists.

        Since the databricks SQL API returns an exception if the table does not exist
        we look out for the relevant exception.
        :param table: Table that may or may not exist
        :return: True if the table exists, false if it does not
        """
        from databricks.sql.exc import ServerOperationError

        try:
            self.hook.run(
                f"DESCRIBE TABLE {table.name}", handler=lambda cur: cur.fetchall_arrow().to_pandas()
            )
            return True
        except ServerOperationError as s:
            if "Table or view not found" in s.message:
                return False
            else:
                raise s

    def create_table_using_columns(self, table: BaseTable) -> None:
        """
        Create a SQL table using the table columns provided by the user.

        :param table: The table to be created.
        """
        if not table.columns:
            raise ValueError("To use this method, table.columns must be defined")
        columns = "\n\t".join([f"{c.name} {c.type}," for c in table.columns])
        self.run_sql(
            dedent(
                f"""
            CREATE TABLE {table.name}
            (
                {columns[:-1]}
            );
            """
            )
        )

    def append_table(
        self,
        source_table: BaseTable,
        target_table: BaseTable,
        source_to_target_columns_map: dict[str, str],
    ) -> None:
        """
        Append the source table rows into a destination table.

        :param source_table: Contains the rows to be appended to the target_table
        :param target_table: Contains the destination table in which the rows will be appended
        :param source_to_target_columns_map: Dict of source_table columns names to target_table columns names
        """

        if source_to_target_columns_map:
            warnings.warn(
                'Warning: Databricks does not support "partial" inserts. '
                "You will need to cast all columns if you wish to use this feature"
            )
            append_query = (
                f"INSERT INTO `{self.get_table_qualified_name(target_table)}` "
                f"({','.join(source_to_target_columns_map.keys())}) SELECT "
                f"{','.join(source_to_target_columns_map.values())} "
                f"FROM `{self.get_table_qualified_name(source_table)}`"
            )
        else:
            append_query = (
                f"INSERT INTO `{self.get_table_qualified_name(target_table)}` "
                f"SELECT * FROM `{self.get_table_qualified_name(source_table)}`"
            )

        self.run_sql(append_query)

    def export_table_to_pandas_dataframe(
        self, source_table: BaseTable, select_kwargs: dict | None = None
    ) -> pd.DataFrame:
        """
        Converts a delta table into a pandas dataframe that can be processed locally.

        Please note that this is a local pandas dataframe and not a spark dataframe. Be careful
        of the size of dataframe you return.
        :param source_table: Delta table to convert to dataframe
        :param select_kwargs: Unused in this function
        :return:
        """

        def convert_delta_table_to_df(cur: Cursor) -> pd.DataFrame:
            df = cur.fetchall_arrow().to_pandas()
            return df

        df = self.hook.run(f"SELECT * FROM {source_table.name}", handler=convert_delta_table_to_df)
        return PandasDataframe.from_pandas_df(df)

    def load_pandas_dataframe_to_table(
        self,
        source_dataframe: pd.DataFrame,
        target_table: BaseTable,
        if_exists: LoadExistStrategy = "replace",
        chunk_size: int = DEFAULT_CHUNK_SIZE,
    ) -> None:
        """
        Create a table with the dataframe's contents.
        If the table already exists, append or replace the content, depending on the value of `if_exists`.

        :param source_dataframe: Local or remote filepath
        :param target_table: Table in which the file will be loaded
        :param if_exists: Strategy to be used in case the target table already exists.
        :param chunk_size: Specify the number of rows in each batch to be written at a time.
        """
        self._assert_not_empty_df(source_dataframe)
        from astro.constants import FileType

        with tempfile.TemporaryDirectory() as t:
            # We have to give each dataframe a unique name because delta does not want to upload the same
            # file multiple times.
            file = File(path=t + f"/dataframe_{uuid.uuid4()}.parquet", filetype=FileType.PARQUET)
            file.create_from_dataframe(source_dataframe)
            self.load_file_to_table(input_file=file, output_table=target_table, if_exists=if_exists)

    @staticmethod
    def get_merge_initialization_query(parameters: tuple) -> str:
        """
        Handles database-specific logic to handle constraints
        for Delta. setting constraints is not required for merging tables in Delta so we only return "1"
        """
        return "SELECT 1"

    def merge_table(
        self,
        source_table: BaseTable,
        target_table: BaseTable,
        source_to_target_columns_map: dict[str, str],
        target_conflict_columns: list[str],
        if_conflicts: MergeConflictStrategy = "exception",
    ):
        """
        Merge the source table rows into a destination table.
        The argument `if_conflicts` allows the user to define how to handle conflicts.

        :param source_table: Contains the rows to be merged to the target_table
        :param target_table: Contains the destination table in which the rows will be merged
        :param source_to_target_columns_map: Dict of target_table columns names to source_table columns names
        :param target_conflict_columns: List of cols where we expect to have a conflict while combining
        :param if_conflicts: The strategy to be applied if there are conflicts.
        """
        statement, params = self._build_merge_sql(
            source_table=source_table,
            target_table=target_table,
            source_to_target_columns_map=source_to_target_columns_map,
            target_conflict_columns=target_conflict_columns,
            if_conflicts=if_conflicts,
        )
        self.run_sql(sql=statement, parameters=params)

    def _build_merge_sql(
        self,
        source_table: BaseTable,
        target_table: BaseTable,
        source_to_target_columns_map: dict[str, str],
        target_conflict_columns: list[str],
        if_conflicts: MergeConflictStrategy = "exception",
    ):
        """Build the SQL statement for Merge operation"""
        # TODO: Simplify this function
        source_table_name = source_table.name
        target_table_name = target_table.name

        source_cols = source_to_target_columns_map.keys()
        target_cols = source_to_target_columns_map.values()

        (
            source_table_identifier,
            source_table_param,
        ) = self.get_sqlalchemy_template_table_identifier_and_parameter(source_table, "source_table")

        (
            target_table_identifier,
            target_table_param,
        ) = self.get_sqlalchemy_template_table_identifier_and_parameter(target_table, "target_table")
        merge_target_dict = {
            f"`target_table`.`{x}`": f"`target_table`.`{x}`" for i, x in enumerate(target_conflict_columns)
        }
        merge_source_dict = {
            f"`source_table`.`{x}`": f"`source_table`.`{x}`" for i, x in enumerate(target_conflict_columns)
        }
        merge_clauses = " AND ".join(
            f"{k}={v}" for k, v in zip(merge_target_dict.keys(), merge_source_dict.keys())
        )
        statement = (
            f"merge into {target_table_identifier} as `target_table` "
            f"using {source_table_identifier} as `source_table`"
            f" on {merge_clauses}"
        )

        values_to_check = [target_table_name, source_table_name]
        values_to_check.extend(source_cols)
        values_to_check.extend(target_cols)
        if if_conflicts == "update":
            merge_statement = ",".join(
                [f"target_table.{t} = source_table.{s}" for s, t in source_to_target_columns_map.items()]
            )
            statement += f" when matched then UPDATE SET {merge_statement}"
        target_columns = ",".join(f"target_table.{t}" for t in target_cols)
        append_columns = ",".join(f"source_table.{s}" for s in source_cols)
        statement += f" when not matched then insert({target_columns}) values ({append_columns})"

        params = {
            **merge_target_dict,
            **merge_source_dict,
            "source_table": source_table_param,
            "target_table": target_table_param,
        }
        return statement, params



import logging

log = logging.getLogger(__name__)

try:
    from openlineage.airflow.extractors import TaskMetadata
    from openlineage.airflow.extractors.base import BaseExtractor, OperatorLineage
    from openlineage.airflow.utils import get_job_name
    from openlineage.client.facet import (
        BaseFacet,
        DataQualityMetricsInputDatasetFacet,
        DataSourceDatasetFacet,
        OutputStatisticsOutputDatasetFacet,
        SchemaDatasetFacet,
        SchemaField,
        SourceCodeJobFacet,
        SqlJobFacet,
    )
    from openlineage.client.run import Dataset as OpenlineageDataset
except ImportError:
    logging.debug("openlineage-airflow python dependency is missing")



from airflow.configuration import conf
from airflow.decorators.base import get_unique_task_id
from airflow.models.xcom_arg import XComArg

from astro.sql.operators.append import AppendOperator, append
from astro.sql.operators.cleanup import CleanupOperator, cleanup
from astro.sql.operators.data_validations.check_column import ColumnCheckOperator, check_column
from astro.sql.operators.data_validations.check_table import SQLCheckOperator, check_table
from astro.sql.operators.dataframe import DataframeOperator, dataframe
from astro.sql.operators.drop import DropTableOperator, drop_table
from astro.sql.operators.export_file import ExportFileOperator, export_file
from astro.sql.operators.export_table_to_file import ExportTableToFileOperator, export_table_to_file
from astro.sql.operators.export_to_file import ExportToFileOperator, export_to_file
from astro.sql.operators.load_file import LoadFileOperator, load_file
from astro.sql.operators.merge import MergeOperator, merge
from astro.sql.operators.raw_sql import RawSQLOperator, run_raw_sql
from astro.sql.operators.transform import TransformOperator, transform, transform_file
from astro.table import Metadata, Table

__all__ = [
    "AppendOperator",
    "append",
    "CleanupOperator",
    "cleanup",
    "DataframeOperator",
    "dataframe",
    "DropTableOperator",
    "drop_table",
    "ExportFileOperator",
    "export_file",
    "ExportTableToFileOperator",
    "export_table_to_file",
    "ExportToFileOperator",
    "export_to_file",
    "get_value_list",
    "LoadFileOperator",
    "load_file",
    "MergeOperator",
    "merge",
    "Metadata",
    "run_raw_sql",
    "Table",
    "TransformOperator",
    "transform_file",
    "transform",
    "ColumnCheckOperator",
    "check_column",
    "SQLCheckOperator",
    "check_table",
]


def get_value_list(sql: str, conn_id: str, **kwargs) -> XComArg:
    """
    Execute a sql statement and return the result.
    By default, the response size is less than equal to value of ``max_map_length`` conf.
    You can call a callable handler to alter the response by default it call ``fetchall`` on database result set.


    :param sql: sql query to execute.
        If the sql query will return huge number of row then it can overload the XCOM.
        also, If you are using output of this method to expand a task using dynamic task map then
        it can create lots of parallel task. So it is advisable to limit your sql query statement.
    :param conn_id: Airflow connection id. This connection id will be used to identify the database client
        and connect with it at runtime
    """
    handler = kwargs.get("handler") or (lambda result_set: result_set.fetchall())
    max_map_length = int(conf.get(section="core", key="max_map_length"))
    op_kwargs = {
        "handler": handler,
        "response_limit": max_map_length,
    }
    task_id = kwargs.get("task_id") or get_unique_task_id(
        "get_value_list", dag=kwargs.get("dag"), task_group=kwargs.get("task_group")
    )
    kwargs.update({"task_id": task_id})
    return RawSQLOperator(
        sql=sql, conn_id=conn_id, op_kwargs=op_kwargs, python_callable=(lambda *args: None), **kwargs
    ).output



from __future__ import annotations

import logging
from collections.abc import Iterable, Mapping
from typing import Any, Callable

try:
    from airflow.decorators.base import TaskDecorator
except ImportError:
    from airflow.decorators import _TaskDecorator as TaskDecorator

import airflow
import pandas as pd
from airflow.decorators.base import task_decorator_factory
from sqlalchemy.engine import ResultProxy

if airflow.__version__ >= "2.3":
    from sqlalchemy.engine.row import LegacyRow as SQLAlcRow
else:
    from sqlalchemy.engine.result import RowProxy as SQLAlcRow

from astro import settings
from astro.constants import RunRawSQLResultFormat
from astro.dataframes.pandas import PandasDataframe
from astro.exceptions import IllegalLoadToDatabaseException
from astro.sql.operators.base_decorator import BaseSQLDecoratedOperator
from astro.utils.compat.typing import Context


class RawSQLOperator(BaseSQLDecoratedOperator):
    """
    Given a SQL statement, (optional) tables and a (optional) function, execute the SQL statement
    and apply the function to the results, returning the result of the function.

    Disclaimer: this could potentially trash the XCom Database, depending on the XCom backend used
    and on the SQL statement/function declared by the user.
    """

    def __init__(
        self,
        results_format: RunRawSQLResultFormat | None = None,
        fail_on_empty: bool = False,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)

        # add the results_format and fail_on_empty to the kwargs
        self.results_format = results_format
        self.fail_on_empty = fail_on_empty

    def execute(self, context: Context) -> Any:
        super().execute(context)

        result = self.database_impl.run_sql(sql=self.sql, parameters=self.parameters, handler=self.handler)
        if self.response_size == -1 and not settings.IS_CUSTOM_XCOM_BACKEND:
            logging.warning(
                "Using `run_raw_sql` without `response_size` can result in excessive amount of data being recorded "
                "to the Airflow metadata database, leading to issues to the orchestration of tasks. It is possible to "
                "avoid this problem by either setting `response_size` to a small integer or by using a custom XCom "
                "backend."
            )

        # ToDo: Currently, the handler param in run_sql() method is only used in databricks all other databases are
        # not using it. Which leads to different response types since handler is processed within `run_sql()` for
        # databricks and not for other databases. Also the signature of `run_sql()` in databricks deviates from base.
        # We need to standardise and when we do, we can remove below check as well.
        if self.database_impl.IGNORE_HANDLER_IN_RUN_RAW_SQL:
            return result

        self.handler = self.get_handler()

        if self.handler:
            self.handler = self.get_wrapped_handler(
                fail_on_empty=self.fail_on_empty, conversion_func=self.handler
            )
            # otherwise, call the handler and convert the result to a list
            response = self.handler(result)
            response = self.make_row_serializable(response)
            if 0 <= self.response_limit < len(response):
                raise IllegalLoadToDatabaseException()  # pragma: no cover
            if self.response_size >= 0:
                resp = response[: self.response_size]
                # We do the following is slicing pandas dataframe creates a pd.Dataframe object
                # which isn't serializable
                # TODO: We could add a __getitem__ to PandasDataframe object
                return PandasDataframe.from_pandas_df(resp) if isinstance(resp, pd.DataFrame) else resp
            else:
                return response

        # if no results_format or handler is specified, we don't return results
        return None

    @staticmethod
    def make_row_serializable(rows: Any) -> Any:
        """
        Convert rows to a serializable format
        """
        if not settings.NEED_CUSTOM_SERIALIZATION:
            return rows
        if isinstance(rows, PandasDataframe):
            return rows
        if isinstance(rows, Iterable):
            return [SdkLegacyRow.from_legacy_row(r) if isinstance(r, SQLAlcRow) else r for r in rows]
        return rows

    @staticmethod
    def results_as_list(results: ResultProxy) -> list:
        """
        Convert the result of a SQL query to a list
        """
        data = []
        for result in results.fetchall():
            data.append(result.values())
        return data

    @staticmethod
    def results_as_pandas_dataframe(result: ResultProxy) -> pd.DataFrame:
        """
        Convert the result of a SQL query to a pandas dataframe
        """
        return PandasDataframe(result.fetchall(), columns=result.keys())

    def get_results_format_handler(self, results_format: str):
        return getattr(self, f"results_as_{results_format}")

    def get_handler(self):
        # if a user specifies a results_format, we will return the results in that format
        # note that it takes precedence over the handler
        if self.results_format:
            self.handler = self.get_results_format_handler(results_format=self.results_format)
            logging.info("Provided 'handler' will be overridden when 'results_format' is given.")
        if self.handler:
            self.handler = self.get_wrapped_handler(
                fail_on_empty=self.fail_on_empty, conversion_func=self.handler
            )
        return self.handler

    @staticmethod
    def get_wrapped_handler(fail_on_empty: bool, conversion_func: Callable):
        # if fail_on_empty is set to False, wrap in a try/except block
        # this is because the conversion function will fail if the result is empty
        # due to sqlalchemy failing when there are no results
        if not fail_on_empty:

            def handle_exceptions(result):
                try:
                    return conversion_func(result)
                except Exception as e:  # skipcq: PYL-W0703
                    logging.info("Exception %s handled since 'fail_on_empty' parameter was passed", str(e))
                    return None

            return handle_exceptions

        # otherwise, just return the result
        return conversion_func


class SdkLegacyRow(SQLAlcRow):
    version: int = 1

    def serialize(self):
        return {"key_map": self._keymap, "key_style": self._key_style, "data": self._data}

    @staticmethod
    def deserialize(data: dict, version: int):  # skipcq: PYL-W0613
        return SdkLegacyRow(None, None, data["key_map"], data["key_style"], data["data"])

    @staticmethod
    def from_legacy_row(obj):
        return SdkLegacyRow(None, None, obj._keymap, obj._key_style, obj._data)  # skipcq: PYL-W0212


def run_raw_sql(
    python_callable: Callable | None = None,
    conn_id: str = "",
    parameters: Mapping | Iterable | None = None,
    database: str | None = None,
    schema: str | None = None,
    handler: Callable | None = None,
    results_format: RunRawSQLResultFormat | None = None,
    fail_on_empty: bool = False,
    response_size: int = settings.RAW_SQL_MAX_RESPONSE_SIZE,
    **kwargs: Any,
) -> TaskDecorator:
    """
    Given a python function that returns a SQL statement and (optional) tables, execute the SQL statement and output
    the result into a SQL table.

    Use this function as a decorator like so:


    .. code-block:: python

      @run_raw_sql
      def my_sql_statement(table1: Table) -> Table:
          return "DROP TABLE {{table1}}"

    In this example, by identifying parameters as ``Table`` objects, astro knows to automatically convert those
    objects into tables (if they are, for example, a dataframe). Any type besides table will lead astro to assume
    you do not want the parameter converted.

    Please note that the ``run_raw_sql`` function will not create a temporary table. It will either return the
    result of a provided ``handler`` function or it will not return anything at all.


    :param python_callable: This parameter is filled in automatically when you use the transform function as a
        decorator. This is where the python function gets passed to the wrapping function
    :param conn_id: Connection ID for the database you want to connect to.
        If you do not pass in a value for this object we can infer the connection ID from the first table
        passed into the python_callable function. (required if there are no table arguments)
    :param parameters: parameters to pass into the SQL query
    :param database: Database within the SQL instance you want to access. If left blank we will default to the
        table.metatadata.database in the first Table passed to the function
        (required if there are no table arguments)
    :param schema: Schema within the SQL instance you want to access. If left blank we will default to the
        table.metatadata.schema in the first Table passed to the function
        (required if there are no table arguments)
    :param handler: Handler function to process the result of the SQL query. For more information please consult
        https://docs.sqlalchemy.org/en/14/core/connections.html#sqlalchemy.engine.Result
    :param results_format: Format of the results returned by the SQL query. Overrides the handler, if provided.
    :param fail_on_empty: If True and a results_format is provided, raises an exception if the query returns no results.
    :param response_size: Used to trim the responses returned to avoid trashing the Airflow DB.
        The default value is -1, which means the response is not changed. Otherwise, if the response is a list,
        returns up to the desired amount of items. If the response is a string, trims it to the desired size.
    :param kwargs:
    :return: By default returns None unless there is a handler function,
        in which case returns the result of the handler
    """

    kwargs.update(
        {
            "conn_id": conn_id,
            "parameters": parameters,
            "database": database,
            "schema": schema,
            "handler": handler,
            "results_format": results_format,
            "fail_on_empty": fail_on_empty,
            "response_size": response_size,
        }
    )
    return task_decorator_factory(
        python_callable=python_callable,
        multiple_outputs=False,
        decorated_operator_class=RawSQLOperator,
        **kwargs,
    )



from __future__ import annotations

import sys
from enum import Enum

# typing.Literal was only introduced in Python 3.8, and we support Python 3.7
if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

DEFAULT_SCHEMA = "tmp_astro"
DEFAULT_CHUNK_SIZE = 1000000
PYPI_PROJECT_NAME = "astro-sdk-python"


class FileLocation(Enum):
    # [START filelocation]
    LOCAL = "local"
    HTTP = "http"
    HTTPS = "https"
    GS = "gs"  # Google Cloud Storage
    GOOGLE_DRIVE = "gdrive"
    S3 = "s3"  # Amazon S3
    WASB = "wasb"  # Azure Blob Storage
    WASBS = "wasbs"  # Azure Blob Storage
    SFTP = "sftp"  # Remote file location
    FTP = "ftp"  # Remote file location
    # [END filelocation]

    def __str__(self) -> str:
        return self.value


class FileType(Enum):
    # [START filetypes]
    CSV = "csv"
    JSON = "json"
    NDJSON = "ndjson"
    PARQUET = "parquet"
    # [END filetypes]

    def __str__(self) -> str:
        return self.value


class Database(Enum):
    # [START database]
    POSTGRES = "postgres"
    POSTGRESQL = "postgres"
    SQLITE = "sqlite"
    DELTA = "delta"
    BIGQUERY = "bigquery"
    SNOWFLAKE = "snowflake"
    REDSHIFT = "redshift"
    MSSQL = "mssql"
    # [END database]

    def __str__(self) -> str:
        return self.value


class DatabricksLoadMode(str, Enum):
    AUTOLOADER = "autoloader"
    COPY_INTO = "copy_into"


SUPPORTED_FILE_LOCATIONS = [const.value for const in FileLocation]
SUPPORTED_FILE_TYPES = [const.value for const in FileType]
SUPPORTED_DATABASES = [const.value for const in Database]

LoadExistStrategy = Literal["replace", "append"]

ExportExistsStrategy = Literal["replace", "exception"]

# TODO: check how snowflake names these
MergeConflictStrategy = Literal["ignore", "update", "exception"]

ColumnCapitalization = Literal["upper", "lower", "original"]

RunRawSQLResultFormat = Literal["list", "pandas_dataframe"]



"""Postgres database implementation."""
from __future__ import annotations

import io
from contextlib import closing

import pandas as pd
import sqlalchemy
from airflow.providers.postgres.hooks.postgres import PostgresHook
from psycopg2 import sql as postgres_sql

from astro.constants import DEFAULT_CHUNK_SIZE, LoadExistStrategy, MergeConflictStrategy
from astro.databases.base import BaseDatabase
from astro.files import File
from astro.options import LoadOptions
from astro.settings import POSTGRES_SCHEMA
from astro.table import BaseTable, Metadata
from astro.utils.compat.functools import cached_property

DEFAULT_CONN_ID = PostgresHook.default_conn_name


class PostgresDatabase(BaseDatabase):
    """
    Handle interactions with Postgres databases. If this class is successful, we should not have any Postgres-specific
    logic in other parts of our code-base.
    """

    DEFAULT_SCHEMA = POSTGRES_SCHEMA
    illegal_column_name_chars: list[str] = ["."]
    illegal_column_name_chars_replacement: list[str] = ["_"]

    def __init__(
        self,
        conn_id: str = DEFAULT_CONN_ID,
        table: BaseTable | None = None,
        load_options: LoadOptions | None = None,
    ):
        super().__init__(conn_id)
        self.table = table
        self.load_options = load_options

    @property
    def sql_type(self) -> str:
        return "postgresql"

    @cached_property
    def hook(self) -> PostgresHook:
        """Retrieve Airflow hook to interface with the Postgres database."""
        conn = PostgresHook(postgres_conn_id=self.conn_id).get_connection(self.conn_id)
        kwargs = {}
        if (conn.schema is None) and (self.table and self.table.metadata and self.table.metadata.database):
            kwargs.update({"database": self.table.metadata.database})
        return PostgresHook(postgres_conn_id=self.conn_id, **kwargs)

    @property
    def default_metadata(self) -> Metadata:
        """Fill in default metadata values for table objects addressing Postgres databases.

        Currently, Schema is not being fetched from airflow connection for Postgres because, in Postgres,
        databases and schema are different concepts: https://www.postgresql.org/docs/current/ddl-schemas.html

        The PostgresHook only exposes schema:
        :external+airflow-postgres:py:class:`airflow.providers.postgres.hooks.postgres.PostgresHook`

        However, implementation-wise, it seems that if the PostgresHook receives a schema during
        initialization, but it uses it as a database in the connection to Postgres:
        https://github.com/apache/airflow/blob/main/airflow/providers/postgres/hooks/postgres.py#L96
        """
        # TODO: Change airflow PostgresHook to fetch database and schema separately
        database = self.hook.get_connection(self.conn_id).schema
        return Metadata(database=database, schema=self.DEFAULT_SCHEMA)  # type: ignore

    def schema_exists(self, schema) -> bool:
        """
        Checks if a schema exists in the database

        :param schema: DB Schema - a namespace that contains named objects like (tables, functions, etc)
        """
        schema_result = self.hook.run(
            "SELECT schema_name FROM information_schema.schemata WHERE lower(schema_name) = lower(%(schema_name)s);",
            parameters={"schema_name": schema.lower()},
            handler=lambda x: [y[0] for y in x.fetchall()],
        )
        return len(schema_result) > 0

    # Require skipcq because method overriding we need param chunk_size
    def load_pandas_dataframe_to_table(
        self,
        source_dataframe: pd.DataFrame,
        target_table: BaseTable,
        if_exists: LoadExistStrategy = "replace",
        chunk_size: int = DEFAULT_CHUNK_SIZE,
    ) -> None:  # skipcq PYL-W0613
        """
        Create a table with the dataframe's contents.
        If the table already exists, append or replace the content, depending on the value of `if_exists`.

        :param source_dataframe: Local or remote filepath
        :param target_table: Table in which the file will be loaded
        :param if_exists: Strategy to be used in case the target table already exists.
        :param chunk_size: Specify the number of rows in each batch to be written at a time.
        """
        self._assert_not_empty_df(source_dataframe)

        self.create_schema_if_needed(target_table.metadata.schema)
        if not self.table_exists(table=target_table) or if_exists == "replace":
            self.create_table(table=target_table, dataframe=source_dataframe)

        output_buffer = io.StringIO()
        source_dataframe.to_csv(output_buffer, sep=",", header=True, index=False)
        output_buffer.seek(0)
        table_name = self.get_table_qualified_name(target_table)
        postgres_conn = self.hook.get_conn()
        with closing(postgres_conn) as conn, closing(conn.cursor()) as cur:
            cur.copy_expert(
                f"COPY {table_name} FROM STDIN DELIMITER ',' CSV HEADER;",
                output_buffer,
            )
            conn.commit()

    @staticmethod
    def get_table_qualified_name(table: BaseTable) -> str:  # skipcq: PYL-R0201
        """
        Return table qualified name. This is Database-specific.
        For instance, in Sqlite this is the table name. In Snowflake, however, it is the database, schema and table

        :param table: The table we want to retrieve the qualified name for.
        """
        # Initially this method belonged to the Table class.
        # However, in order to have an agnostic table class implementation,
        # we are keeping all methods which vary depending on the database within the Database class.
        if table.metadata and table.metadata.schema:
            qualified_name = f"{table.metadata.schema.lower()}.{table.name}"  # type: ignore
        else:
            qualified_name = table.name
        return qualified_name

    def table_exists(self, table: BaseTable) -> bool:
        """
        Check if a table exists in the database

        :param table: Details of the table we want to check that exists
        """
        _schema = table.metadata.schema
        # when creating schemas they are created in a lowercase even when we have a schema in uppercase.
        # while checking for schema there in no lowercase applied in 'has_table()' which leads to table not found.
        # Added 'schema.lower()' to make sure we search for schema in lowercase to match the creation lowercase.
        schema = _schema.lower() if _schema else _schema

        inspector = sqlalchemy.inspect(self.sqlalchemy_engine)
        return bool(inspector.dialect.has_table(self.connection, table.name, schema))

    def merge_table(
        self,
        source_table: BaseTable,
        target_table: BaseTable,
        source_to_target_columns_map: dict[str, str],
        target_conflict_columns: list[str],
        if_conflicts: MergeConflictStrategy = "exception",
    ) -> None:
        """
        Merge the source table rows into a destination table.
        The argument `if_conflicts` allows the user to define how to handle conflicts.

        :param source_table: Contains the rows to be merged to the target_table
        :param target_table: Contains the destination table in which the rows will be merged
        :param source_to_target_columns_map: Dict of target_table columns names to source_table columns names
        :param target_conflict_columns: List of cols where we expect to have a conflict while combining
        :param if_conflicts: The strategy to be applied if there are conflicts.
        """

        def identifier_args(table: BaseTable):
            schema = table.metadata.schema
            return (schema, table.name) if schema else (table.name,)

        statement = (
            "INSERT INTO {target_table} ({target_columns}) SELECT {source_columns} FROM {source_table}"
        )

        source_columns = list(source_to_target_columns_map.keys())
        target_columns = list(source_to_target_columns_map.values())

        if if_conflicts == "ignore":
            statement += " ON CONFLICT ({target_conflict_columns}) DO NOTHING"
        elif if_conflicts == "update":
            statement += " ON CONFLICT ({target_conflict_columns}) DO UPDATE SET {update_statements}"

        source_column_names = [postgres_sql.Identifier(col) for col in source_columns]
        target_column_names = [postgres_sql.Identifier(col) for col in target_columns]
        update_statements = [
            postgres_sql.SQL("{col_name}=EXCLUDED.{col_name}").format(col_name=col_name)
            for col_name in target_column_names
        ]

        query = postgres_sql.SQL(statement).format(
            target_columns=postgres_sql.SQL(",").join(target_column_names),
            target_table=postgres_sql.Identifier(*identifier_args(target_table)),
            source_columns=postgres_sql.SQL(",").join(source_column_names),
            source_table=postgres_sql.Identifier(*identifier_args(source_table)),
            update_statements=postgres_sql.SQL(",").join(update_statements),
            target_conflict_columns=postgres_sql.SQL(",").join(
                [postgres_sql.Identifier(x) for x in target_conflict_columns]
            ),
        )

        sql = query.as_string(self.hook.get_conn())
        self.run_sql(sql=sql)

    @staticmethod
    def get_dataframe_from_file(file: File):  # skipcq: PYL-W0613
        """
        Get pandas dataframe file

        :param file: File path and conn_id for object stores
        """
        return file.export_to_dataframe_via_byte_stream()

    def openlineage_dataset_name(self, table: BaseTable) -> str:
        """
        Returns the open lineage dataset name as per
        https://github.com/OpenLineage/OpenLineage/blob/main/spec/Naming.md
        Example: schema_name.table_name
        """
        schema = self.hook.get_connection(self.conn_id).schema or "public"
        return f"{schema}.{table.name}"

    def openlineage_dataset_namespace(self) -> str:
        """
        Returns the open lineage dataset namespace as per
        https://github.com/OpenLineage/OpenLineage/blob/main/spec/Naming.md
        Example: postgresql://localhost:5432
        """
        conn = self.hook.get_connection(self.conn_id)
        return f"{self.sql_type}://{conn.host}:{conn.port}"

    def openlineage_dataset_uri(self, table: BaseTable) -> str:
        """
        Returns the open lineage dataset uri as per
        https://github.com/OpenLineage/OpenLineage/blob/main/spec/Naming.md
        """
        return f"{self.openlineage_dataset_namespace()}/{self.openlineage_dataset_name(table=table)}"



from typing import TYPE_CHECKING, Any

from airflow.exceptions import AirflowException
from airflow.models.xcom import BaseXCom

from astro.custom_backend.serializer import deserialize, serialize

if TYPE_CHECKING:
    from airflow.models.xcom import XCom


class AstroCustomXcomBackend(BaseXCom):
    """
    The Astro custom xcom backend is an xcom custom backend wrapper that handles serialization and deserialization
    of command data types used in astro. The astro library overrides the ``TaskInstance.XCom`` object with this wrapper.

    Once the wrapper has properly serialized the objects, the json-safe dictionaries are then passed to the original
    XCom object. This ensures that custom backends still work and there are no breaking changes to airflow users.
    """

    @staticmethod
    def serialize_value(value: Any, **kwargs) -> Any:  # type: ignore[override]
        """
        Serialize the return value of a taskinstance and then pass it to the XCom.serialize_value function.
        :param value: object to be serialized.
        :param kwargs:
        :return:
        """
        from astro.settings import DATAFRAME_STORAGE_CONN_ID, STORE_DATA_LOCAL_DEV

        if DATAFRAME_STORAGE_CONN_ID or STORE_DATA_LOCAL_DEV:
            value = serialize(value)
        else:
            raise AirflowException(
                "Since you have not provided a remote object storage conn_id, we would need to store your dataframes"
                " in the Metadata DB. This does not scale well and can cause degradation to "
                "your airflow DB. Please set the AIRFLOW__ASTRO_SDK__XCOM_STORAGE_CONN_ID and "
                "AIRFLOW__ASTRO_SDK__XCOM_STORAGE_URL variables "
                "(or you can set AIRFLOW__ASTRO_SDK__STORE_DATA_LOCAL_DEV for local development)"
            )
        return BaseXCom.serialize_value(value, **kwargs)

    @staticmethod
    def deserialize_value(result: "XCom") -> Any:
        """
        Deserializing the result of a xcom_pull before passing th result to the next task.
        :param result:
        :return:
        """
        result = BaseXCom.deserialize_value(result)
        return deserialize(result)



from __future__ import annotations

from typing import Any

import pandas as pd
from airflow.decorators.base import get_unique_task_id
from airflow.models.xcom_arg import XComArg

from astro.airflow.datasets import kwargs_with_datasets
from astro.constants import ExportExistsStrategy
from astro.databases import create_database
from astro.files import File
from astro.sql.operators.base_operator import AstroSQLBaseOperator
from astro.table import BaseTable, Table
from astro.utils.compat.typing import Context


class ExportToFileOperator(AstroSQLBaseOperator):
    """Write SQL table to csv/parquet on local/S3/GCS.

    :param input_data: Table to convert to file
    :param output_file: File object containing the path to the file and connection id.
    :param if_exists: Overwrite file if exists. Default False.
    """

    template_fields = ("input_data", "output_file")

    def __init__(
        self,
        input_data: BaseTable | pd.DataFrame,
        output_file: File,
        if_exists: ExportExistsStrategy = "exception",
        **kwargs,
    ) -> None:
        self.output_file = output_file
        self.input_data = input_data
        self.if_exists = if_exists
        self.kwargs = kwargs
        datasets = {"output_datasets": self.output_file}
        if isinstance(input_data, Table):
            datasets["input_datasets"] = input_data
        super().__init__(**kwargs_with_datasets(kwargs=kwargs, **datasets))

    def execute(self, context: Context) -> File:  # skipcq PYL-W0613
        """Write SQL table to csv/parquet on local/S3/GCS.

        Infers SQL database type based on connection.
        """
        # Infer db type from `input_conn_id`.
        if isinstance(self.input_data, BaseTable):
            database = create_database(self.input_data.conn_id, table=self.input_data)
            self.input_data = database.populate_table_metadata(self.input_data)
            df = database.export_table_to_pandas_dataframe(self.input_data)
        elif isinstance(self.input_data, pd.DataFrame):
            df = self.input_data
        else:
            raise ValueError(f"Expected input_table to be Table or dataframe. Got {type(self.input_data)}")
        # Write file if overwrite == True or if file doesn't exist.
        if self.if_exists == "replace" or not self.output_file.exists():
            self.output_file.create_from_dataframe(df, store_as_dataframe=False)
            return self.output_file
        else:
            raise FileExistsError(f"{self.output_file.path} file already exists.")

    def get_openlineage_facets_on_complete(self, task_instance):  # skipcq: PYL-W0613
        """
        Collect the input, output, job and run facets for export file operator
        """

        from astro.lineage import (
            BaseFacet,
            DataQualityMetricsInputDatasetFacet,
            DataSourceDatasetFacet,
            OpenlineageDataset,
            OperatorLineage,
            OutputStatisticsOutputDatasetFacet,
            SchemaDatasetFacet,
            SchemaField,
        )
        from astro.lineage.facets import ExportFileFacet

        input_dataset: list[OpenlineageDataset] = []
        if isinstance(self.input_data, BaseTable) and self.input_data.openlineage_emit_temp_table_event():
            input_dataset = [
                OpenlineageDataset(
                    namespace=self.input_data.openlineage_dataset_namespace(),
                    name=self.input_data.openlineage_dataset_name(),
                    facets={
                        "dataSource": DataSourceDatasetFacet(
                            name=self.input_data.openlineage_dataset_name(),
                            uri=self.input_data.openlineage_dataset_uri(),
                        ),
                        "schema": SchemaDatasetFacet(
                            fields=[
                                SchemaField(
                                    name=self.input_data.metadata.schema,
                                    type=self.input_data.metadata.database,
                                )
                            ]
                        ),
                        "dataQualityMetrics": DataQualityMetricsInputDatasetFacet(
                            rowCount=self.input_data.row_count, columnMetrics={}
                        ),
                    },
                )
            ]
        output_uri = (
            f"{self.output_file.openlineage_dataset_namespace}{self.output_file.openlineage_dataset_name}"
        )
        output_dataset = [
            OpenlineageDataset(
                namespace=self.output_file.openlineage_dataset_namespace,
                name=self.output_file.openlineage_dataset_name,
                facets={
                    "output_file_facet": ExportFileFacet(
                        filepath=self.output_file.path,
                        file_size=self.output_file.size,
                        file_type=self.output_file.type.name,
                        if_exists=self.if_exists,
                    ),
                    "dataSource": DataSourceDatasetFacet(name=self.output_file.path, uri=output_uri),
                    "outputStatistics": OutputStatisticsOutputDatasetFacet(
                        rowCount=self.input_data.row_count
                        if isinstance(self.input_data, BaseTable)
                        else len(self.input_data),
                        size=self.output_file.size,
                    ),
                },
            )
        ]

        run_facets: dict[str, BaseFacet] = {}
        job_facets: dict[str, BaseFacet] = {}

        return OperatorLineage(
            inputs=input_dataset, outputs=output_dataset, run_facets=run_facets, job_facets=job_facets
        )


def export_to_file(
    input_data: BaseTable | pd.DataFrame,
    output_file: File,
    if_exists: ExportExistsStrategy = "exception",
    task_id: str | None = None,
    **kwargs: Any,
) -> XComArg:
    """Convert ExportToFileOperator into a function. Returns XComArg.

    Returns an XComArg object of type File which matches the output_file parameter.

    This will allow users to perform further actions with the exported file.

    e.g.:

    .. code-block:: python

      with sample_dag:
          table = aql.load_file(input_file=File(path=data_path), output_table=test_table)
          exported_file = aql.export_file(
              input_data=table,
              output_file=File(path="/tmp/saved_df.csv"),
              if_exists="replace",
          )
          res_df = aql.load_file(input_file=exported_file)


    :param output_file: Path and conn_id
    :param input_data: Input table / dataframe
    :param if_exists: Overwrite file if exists. Default "exception"
    :param task_id: task id, optional
    """

    task_id = task_id or get_unique_task_id("export_to_file")

    return ExportToFileOperator(
        task_id=task_id,
        output_file=output_file,
        input_data=input_data,
        if_exists=if_exists,
        **kwargs,
    ).output



from __future__ import annotations

import glob
import os
import pathlib
from urllib.parse import urlparse

from astro.constants import FileLocation
from astro.files.locations.base import BaseFileLocation


class LocalLocation(BaseFileLocation):
    """Handler Local file path operations"""

    location_type = FileLocation.LOCAL

    @property
    def paths(self) -> list[str]:
        """Resolve local filepath"""
        url = urlparse(self.path)
        path_object = pathlib.Path(url.path)
        if path_object.is_dir():
            paths = [str(filepath) for filepath in path_object.rglob("*")]
        else:
            paths = glob.glob(url.path)
        return paths

    def validate_conn(self):
        """Override as conn_id is not always required for local location."""

    @property
    def size(self) -> int:
        """Return the size in bytes of the given file.

        :return: File size in bytes
        """
        path = pathlib.Path(self.path)
        return os.path.getsize(path)

    @property
    def openlineage_dataset_namespace(self) -> str:
        """
        Returns the open lineage dataset namespace as per
        https://github.com/OpenLineage/OpenLineage/blob/main/spec/Naming.md
        """
        return os.path.basename(self.path)

    @property
    def openlineage_dataset_name(self) -> str:
        """
        Returns the open lineage dataset name as per
        https://github.com/OpenLineage/OpenLineage/blob/main/spec/Naming.md
        """
        return urlparse(self.path).path



from __future__ import annotations

import logging
import warnings
from abc import ABC
from typing import TYPE_CHECKING, Any, Callable, Mapping

import pandas as pd
import sqlalchemy
from airflow.hooks.dbapi import DbApiHook
from pandas.io.sql import SQLDatabase
from sqlalchemy import column, insert, select

from astro.dataframes.pandas import PandasDataframe

if TYPE_CHECKING:  # pragma: no cover
    from sqlalchemy.engine.cursor import CursorResult

from sqlalchemy.sql import ClauseElement
from sqlalchemy.sql.elements import ColumnClause
from sqlalchemy.sql.schema import Table as SqlaTable

from astro.constants import (
    DEFAULT_CHUNK_SIZE,
    ColumnCapitalization,
    ExportExistsStrategy,
    FileLocation,
    FileType,
    LoadExistStrategy,
    MergeConflictStrategy,
)
from astro.exceptions import DatabaseCustomError, NonExistentTableException
from astro.files import File, resolve_file_path_pattern
from astro.files.types import create_file_type
from astro.files.types.base import FileType as FileTypeConstants
from astro.options import LoadOptions
from astro.settings import LOAD_FILE_ENABLE_NATIVE_FALLBACK, LOAD_TABLE_AUTODETECT_ROWS_COUNT, SCHEMA
from astro.table import BaseTable, Metadata
from astro.utils.compat.functools import cached_property


class BaseDatabase(ABC):
    """
    Base class to represent all the Database interactions.

    The goal is to be able to support new databases by adding
    a new module to the `astro/databases` directory, without the need of
    changing other modules and classes.

    The exception is if the Airflow connection type does not match the new
    Database module name. In that case, we should update the dictionary
    `CUSTOM_CONN_TYPE_TO_MODULE_PATH` available at `astro/databases/__init__.py`.
    """

    _create_schema_statement: str = "CREATE SCHEMA IF NOT EXISTS {}"
    _drop_table_statement: str = "DROP TABLE IF EXISTS {}"
    _create_table_statement: str = "CREATE TABLE IF NOT EXISTS {} AS {}"
    # Used to normalize the ndjson when appending fields in nested fields.
    # Example -
    #   ndjson - {'a': {'b': 'val'}}
    #   the col names generated is 'a.b'. char '.' maybe an illegal char in some db's col name.
    # Contains the illegal char and there replacement, where the value in
    # illegal_column_name_chars[0] will be replaced by value in illegal_column_name_chars_replacement[0]
    illegal_column_name_chars: list[str] = []
    illegal_column_name_chars_replacement: list[str] = []
    # In run_raw_sql operator decides if we want to return results directly or process them by handler provided
    IGNORE_HANDLER_IN_RUN_RAW_SQL: bool = False
    NATIVE_PATHS: dict[Any, Any] = {}
    DEFAULT_SCHEMA = SCHEMA
    NATIVE_LOAD_EXCEPTIONS: Any = DatabaseCustomError
    NATIVE_AUTODETECT_SCHEMA_CONFIG: Mapping[FileLocation, Mapping[str, list[FileType] | Callable]] = {}
    FILE_PATTERN_BASED_AUTODETECT_SCHEMA_SUPPORTED: set[FileLocation] = set()

    def __init__(
        self,
        conn_id: str,
        table: BaseTable | None = None,  # skipcq: PYL-W0613
        load_options: LoadOptions | None = None,
    ):
        self.conn_id = conn_id
        self.sql: str | ClauseElement = ""
        self.load_options = load_options

    def __repr__(self):
        return f'{self.__class__.__name__}(conn_id="{self.conn_id})'

    @property
    def sql_type(self):
        raise NotImplementedError

    @cached_property
    def hook(self) -> DbApiHook:
        """Return an instance of the database-specific Airflow hook."""
        raise NotImplementedError

    @property
    def connection(self) -> sqlalchemy.engine.base.Connection:
        """Return a Sqlalchemy connection object for the given database."""
        return self.sqlalchemy_engine.connect()

    @cached_property
    def sqlalchemy_engine(self) -> sqlalchemy.engine.base.Engine:
        """Return Sqlalchemy engine."""
        return self.hook.get_sqlalchemy_engine()  # type: ignore[no-any-return]

    def run_sql(
        self,
        sql: str | ClauseElement = "",
        parameters: dict | None = None,
        **kwargs,
    ) -> CursorResult:
        """
        Return the results to running a SQL statement.

        Whenever possible, this method should be implemented using Airflow Hooks,
        since this will simplify the integration with Async operators.

        :param sql: Contains SQL query to be run against database
        :param parameters: Optional parameters to be used to render the query
        :param autocommit: Optional autocommit flag
        """
        if parameters is None:
            parameters = {}

        if "sql_statement" in kwargs:  # pragma: no cover
            warnings.warn(
                "`sql_statement` is deprecated and will be removed in future release"
                "Please use  `sql` param instead.",
                DeprecationWarning,
                stacklevel=2,
            )
            sql = kwargs.get("sql_statement")  # type: ignore

        # We need to autocommit=True to make sure the query runs. This is done exclusively for SnowflakeDatabase's
        # truncate method to reflect changes.
        if isinstance(sql, str):
            result = self.connection.execute(
                sqlalchemy.text(sql).execution_options(autocommit=True), parameters
            )
        else:
            result = self.connection.execute(sql, parameters)
        return result

    def columns_exist(self, table: BaseTable, columns: list[str]) -> bool:
        """
        Check that a list of columns exist in the given table.

        :param table: The table to check in.
        :param columns: The columns to check.

        :returns: whether the columns exist in the table or not.
        """
        sqla_table = self.get_sqla_table(table)
        return all(
            any(sqla_column.name == column for sqla_column in sqla_table.columns) for column in columns
        )

    def table_exists(self, table: BaseTable) -> bool:
        """
        Check if a table exists in the database.

        :param table: Details of the table we want to check that exists
        """
        table_qualified_name = self.get_table_qualified_name(table)
        inspector = sqlalchemy.inspect(self.sqlalchemy_engine)
        return bool(inspector.dialect.has_table(self.connection, table_qualified_name))

    # ---------------------------------------------------------
    # Table metadata
    # ---------------------------------------------------------
    @staticmethod
    def get_merge_initialization_query(parameters: tuple) -> str:
        """
        Handles database-specific logic to handle constraints, keeping
        it agnostic to database.
        """
        constraints = ",".join(parameters)
        sql = "ALTER TABLE {{table}} ADD CONSTRAINT airflow UNIQUE (%s)" % constraints
        return sql

    @staticmethod
    def get_table_qualified_name(table: BaseTable) -> str:  # skipcq: PYL-R0201
        """
        Return table qualified name. This is Database-specific.
        For instance, in Sqlite this is the table name. In Snowflake, however, it is the database, schema and table

        :param table: The table we want to retrieve the qualified name for.
        """
        # Initially this method belonged to the Table class.
        # However, in order to have an agnostic table class implementation,
        # we are keeping all methods which vary depending on the database within the Database class.
        if table.metadata and table.metadata.schema:
            qualified_name = f"{table.metadata.schema}.{table.name}"
        else:
            qualified_name = table.name
        return qualified_name

    @property
    def default_metadata(self) -> Metadata:
        """
        Extract the metadata available within the Airflow connection associated with
        self.conn_id.

        :return: a Metadata instance
        """
        raise NotImplementedError

    def populate_table_metadata(self, table: BaseTable) -> BaseTable:
        """
        Given a table, check if the table has metadata.
        If the metadata is missing, and the database has metadata, assign it to the table.
        If the table schema was not defined by the end, retrieve the user-defined schema.
        This method performs the changes in-place and also returns the table.

        :param table: Table to potentially have their metadata changed
        :return table: Return the modified table
        """
        if table.metadata and table.metadata.is_empty() and self.default_metadata:
            table.metadata = self.default_metadata
        if not table.metadata.schema:
            table.metadata.schema = self.DEFAULT_SCHEMA
        return table

    # ---------------------------------------------------------
    # Table creation & deletion methods
    # ---------------------------------------------------------
    def create_table_using_columns(self, table: BaseTable) -> None:
        """
        Create a SQL table using the table columns.

        :param table: The table to be created.
        """
        if not table.columns:
            raise ValueError("To use this method, table.columns must be defined")

        metadata = table.sqlalchemy_metadata
        sqlalchemy_table = sqlalchemy.Table(table.name, metadata, *table.columns)
        metadata.create_all(self.sqlalchemy_engine, tables=[sqlalchemy_table])

    def create_table_using_native_schema_autodetection(
        self,
        table: BaseTable,
        file: File,
    ) -> None:
        """
        Create a SQL table, automatically inferring the schema using the given file via native database support.

        :param table: The table to be created.
        :param file: File used to infer the new table columns.
        """
        raise NotImplementedError("Missing implementation of native schema autodetection.")

    def create_table_using_schema_autodetection(
        self,
        table: BaseTable,
        file: File | None = None,
        dataframe: pd.DataFrame | None = None,
        columns_names_capitalization: ColumnCapitalization = "original",  # skipcq
    ) -> None:
        """
        Create a SQL table, automatically inferring the schema using the given file.

        :param table: The table to be created.
        :param file: File used to infer the new table columns.
        :param dataframe: Dataframe used to infer the new table columns if there is no file
        :param columns_names_capitalization: determines whether to convert all columns to lowercase/uppercase
            in the resulting dataframe
        """
        if file is None:
            if dataframe is None:
                raise ValueError(
                    "File or Dataframe is required for creating table using schema autodetection"
                )
            source_dataframe = dataframe
        else:
            source_dataframe = file.export_to_dataframe(nrows=LOAD_TABLE_AUTODETECT_ROWS_COUNT)

        db = SQLDatabase(engine=self.sqlalchemy_engine)
        db.prep_table(
            source_dataframe,
            table.name.lower(),
            schema=table.metadata.schema,
            if_exists="replace",
            index=False,
        )

    def is_native_autodetect_schema_available(  # skipcq: PYL-R0201
        self, file: File  # skipcq: PYL-W0613
    ) -> bool:
        """
        Check if native auto detection of schema is available.

        :param file: File used to check the file type of to decide
            whether there is a native auto detection available for it.
        """
        return False

    def create_table(
        self,
        table: BaseTable,
        file: File | None = None,
        dataframe: pd.DataFrame | None = None,
        columns_names_capitalization: ColumnCapitalization = "original",
    ) -> None:
        """
        Create a table either using its explicitly defined columns or inferring
        it's columns from a given file.

        :param table: The table to be created
        :param file: (optional) File used to infer the table columns.
        :param dataframe: (optional) Dataframe used to infer the new table columns if there is no file
        :param columns_names_capitalization: determines whether to convert all columns to lowercase/uppercase
            in the resulting dataframe
        """
        if table.columns:
            self.create_table_using_columns(table)
        elif file and self.is_native_autodetect_schema_available(file):
            self.create_table_using_native_schema_autodetection(table, file)
        else:
            self.create_table_using_schema_autodetection(
                table,
                file=file,
                dataframe=dataframe,
                columns_names_capitalization=columns_names_capitalization,
            )

    def create_table_from_select_statement(
        self,
        statement: str,
        target_table: BaseTable,
        parameters: dict | None = None,
    ) -> None:
        """
        Export the result rows of a query statement into another table.

        :param statement: SQL query statement
        :param target_table: Destination table where results will be recorded.
        :param parameters: (Optional) parameters to be used to render the SQL query
        """
        statement = self._create_table_statement.format(
            self.get_table_qualified_name(target_table), statement
        )
        self.run_sql(statement, parameters)

    def drop_table(self, table: BaseTable) -> None:
        """
        Delete a SQL table, if it exists.

        :param table: The table to be deleted.
        """
        statement = self._drop_table_statement.format(self.get_table_qualified_name(table))
        self.run_sql(statement)

    # ---------------------------------------------------------
    # Table load methods
    # ---------------------------------------------------------

    def create_schema_and_table_if_needed(
        self,
        table: BaseTable,
        file: File,
        normalize_config: dict | None = None,
        columns_names_capitalization: ColumnCapitalization = "original",
        if_exists: LoadExistStrategy = "replace",
        use_native_support: bool = True,
    ):
        """
        Checks if the autodetect schema exists for native support else creates the schema and table
        :param table: Table to create
        :param file: File path and conn_id for object stores
        :param normalize_config: pandas json_normalize params config
        :param columns_names_capitalization:  determines whether to convert all columns to lowercase/uppercase
        :param if_exists:  Overwrite file if exists
        :param use_native_support: Use native support for data transfer if available on the destination
        """
        is_schema_autodetection_supported = self.check_schema_autodetection_is_supported(source_file=file)
        is_file_pattern_based_schema_autodetection_supported = (
            self.check_file_pattern_based_schema_autodetection_is_supported(source_file=file)
        )
        if if_exists == "replace":
            self.drop_table(table)
        if use_native_support and is_schema_autodetection_supported and not file.is_pattern():
            return
        if (
            use_native_support
            and file.is_pattern()
            and is_schema_autodetection_supported
            and is_file_pattern_based_schema_autodetection_supported
        ):
            return

        self.create_schema_if_needed(table.metadata.schema)
        if if_exists == "replace" or not self.table_exists(table):
            files = resolve_file_path_pattern(
                file.path,
                file.conn_id,
                normalize_config=normalize_config,
                filetype=file.type.name,
                load_options=file.load_options,
            )
            self.create_table(
                table,
                # We only use the first file for inferring the table schema
                files[0],
                columns_names_capitalization=columns_names_capitalization,
            )

    def fetch_all_rows(self, table: BaseTable, row_limit: int = -1) -> list:
        """
        Fetches all rows for a table and returns as a list. This is needed because some
        databases have different cursors that require different methods to fetch rows

        :param row_limit: Limit the number of rows returned, by default return all rows.
        :param table: The table metadata needed to fetch the rows
        :return: a list of rows
        """
        statement = f"SELECT * FROM {self.get_table_qualified_name(table)}"
        if row_limit > -1:
            statement = statement + f" LIMIT {row_limit}"
        response = self.run_sql(statement)
        return response.fetchall()  # type: ignore

    def load_file_to_table(
        self,
        input_file: File,
        output_table: BaseTable,
        normalize_config: dict | None = None,
        if_exists: LoadExistStrategy = "replace",
        chunk_size: int = DEFAULT_CHUNK_SIZE,
        use_native_support: bool = True,
        native_support_kwargs: dict | None = None,
        columns_names_capitalization: ColumnCapitalization = "original",
        enable_native_fallback: bool | None = LOAD_FILE_ENABLE_NATIVE_FALLBACK,
        **kwargs,
    ):
        """
        Load content of multiple files in output_table.
        Multiple files are sourced from the file path, which can also be path pattern.

        :param input_file: File path and conn_id for object stores
        :param output_table: Table to create
        :param if_exists: Overwrite file if exists
        :param chunk_size: Specify the number of records in each batch to be written at a time
        :param use_native_support: Use native support for data transfer if available on the destination
        :param normalize_config: pandas json_normalize params config
        :param native_support_kwargs: kwargs to be used by method involved in native support flow
        :param columns_names_capitalization: determines whether to convert all columns to lowercase/uppercase
            in the resulting dataframe
        :param enable_native_fallback: Use enable_native_fallback=True to fall back to default transfer
        """
        normalize_config = normalize_config or {}

        self.create_schema_and_table_if_needed(
            file=input_file,
            table=output_table,
            columns_names_capitalization=columns_names_capitalization,
            if_exists=if_exists,
            normalize_config=normalize_config,
        )

        if use_native_support and self.is_native_load_file_available(
            source_file=input_file, target_table=output_table
        ):

            self.load_file_to_table_natively_with_fallback(
                source_file=input_file,
                target_table=output_table,
                if_exists="append",
                normalize_config=normalize_config,
                native_support_kwargs=native_support_kwargs,
                enable_native_fallback=enable_native_fallback,
                chunk_size=chunk_size,
            )
        else:
            self.load_file_to_table_using_pandas(
                input_file=input_file,
                output_table=output_table,
                normalize_config=normalize_config,
                if_exists="append",
                chunk_size=chunk_size,
            )

    @staticmethod
    def get_dataframe_from_file(file: File):
        """
        Get pandas dataframe file. We need export_to_dataframe() for Biqqery,Snowflake and Redshift except for Postgres.
        For postgres we are overriding this method and using export_to_dataframe_via_byte_stream().
        export_to_dataframe_via_byte_stream copies files in a buffer and then use that buffer to ingest data.
        With this approach we have significant performance boost for postgres.

        :param file: File path and conn_id for object stores
        """

        return file.export_to_dataframe()

    @staticmethod
    def _assert_not_empty_df(df):
        """Raise error if dataframe empty

        param df: A dataframe
        """
        if df.empty:
            raise ValueError("Can't load empty dataframe")

    def load_file_to_table_using_pandas(
        self,
        input_file: File,
        output_table: BaseTable,
        normalize_config: dict | None = None,
        if_exists: LoadExistStrategy = "replace",
        chunk_size: int = DEFAULT_CHUNK_SIZE,
    ):
        logging.info("Loading file(s) with Pandas...")
        input_files = resolve_file_path_pattern(
            input_file.path,
            input_file.conn_id,
            normalize_config=normalize_config,
            filetype=input_file.type.name,
            load_options=input_file.load_options,
        )

        for file in input_files:
            self.load_pandas_dataframe_to_table(
                self.get_dataframe_from_file(file),
                output_table,
                chunk_size=chunk_size,
                if_exists=if_exists,
            )

    def load_file_to_table_natively_with_fallback(
        self,
        source_file: File,
        target_table: BaseTable,
        if_exists: LoadExistStrategy = "replace",
        normalize_config: dict | None = None,
        native_support_kwargs: dict | None = None,
        enable_native_fallback: bool | None = LOAD_FILE_ENABLE_NATIVE_FALLBACK,
        chunk_size: int = DEFAULT_CHUNK_SIZE,
        **kwargs,
    ):
        """
        Load content of a file in output_table.

        :param source_file: File path and conn_id for object stores
        :param target_table: Table to create
        :param if_exists: Overwrite file if exists
        :param chunk_size: Specify the number of records in each batch to be written at a time
        :param native_support_kwargs: kwargs to be used by method involved in native support flow
        :param enable_native_fallback: Use enable_native_fallback=True to fall back to default transfer
        :param normalize_config: pandas json_normalize params config
        """

        try:
            logging.info("Loading file(s) with Native Support...")
            self.load_file_to_table_natively(
                source_file=source_file,
                target_table=target_table,
                if_exists=if_exists,
                native_support_kwargs=native_support_kwargs,
                **kwargs,
            )
        except self.NATIVE_LOAD_EXCEPTIONS as load_exception:  # skipcq: PYL-W0703
            logging.warning(
                "Loading file(s) failed with Native Support.",
                exc_info=True,
            )
            if enable_native_fallback:
                logging.warning("Falling back to Pandas-based load...")
                self.load_file_to_table_using_pandas(
                    input_file=source_file,
                    output_table=target_table,
                    normalize_config=normalize_config,
                    if_exists=if_exists,
                    chunk_size=chunk_size,
                )
            else:
                raise load_exception

    def load_pandas_dataframe_to_table(
        self,
        source_dataframe: pd.DataFrame,
        target_table: BaseTable,
        if_exists: LoadExistStrategy = "replace",
        chunk_size: int = DEFAULT_CHUNK_SIZE,
    ) -> None:
        """
        Create a table with the dataframe's contents.
        If the table already exists, append or replace the content, depending on the value of `if_exists`.

        :param source_dataframe: Local or remote filepath
        :param target_table: Table in which the file will be loaded
        :param if_exists: Strategy to be used in case the target table already exists.
        :param chunk_size: Specify the number of rows in each batch to be written at a time.
        """
        self._assert_not_empty_df(source_dataframe)

        source_dataframe.to_sql(
            self.get_table_qualified_name(target_table),
            con=self.sqlalchemy_engine,
            if_exists=if_exists,
            chunksize=chunk_size,
            method="multi",
            index=False,
        )

    def append_table(
        self,
        source_table: BaseTable,
        target_table: BaseTable,
        source_to_target_columns_map: dict[str, str],
    ) -> None:
        """
        Append the source table rows into a destination table.

        :param source_table: Contains the rows to be appended to the target_table
        :param target_table: Contains the destination table in which the rows will be appended
        :param source_to_target_columns_map: Dict of source_table columns names to target_table columns names
        """
        target_table_sqla = self.get_sqla_table(target_table)
        source_table_sqla = self.get_sqla_table(source_table)

        target_columns: list[ColumnClause]
        source_columns: list[ColumnClause]

        if not source_to_target_columns_map:
            target_columns = [column(col) for col in target_table_sqla.c.keys()]
            source_columns = target_columns
        else:
            target_columns = [column(col) for col in source_to_target_columns_map.values()]
            source_columns = [column(col) for col in source_to_target_columns_map.keys()]

        sel = select(source_columns).select_from(source_table_sqla)
        # TODO: We should fix the following Type Error
        # incompatible type List[ColumnClause[<nothing>]]; expected List[Column[Any]]
        self.sql = insert(target_table_sqla).from_select(target_columns, sel)  # type: ignore[arg-type]
        self.run_sql(sql=self.sql)

    def merge_table(
        self,
        source_table: BaseTable,
        target_table: BaseTable,
        source_to_target_columns_map: dict[str, str],
        target_conflict_columns: list[str],
        if_conflicts: MergeConflictStrategy = "exception",
    ) -> None:
        """
        Merge the source table rows into a destination table.
        The argument `if_conflicts` allows the user to define how to handle conflicts.

        :param source_table: Contains the rows to be merged to the target_table
        :param target_table: Contains the destination table in which the rows will be merged
        :param source_to_target_columns_map: Dict of target_table columns names to source_table columns names
        :param target_conflict_columns: List of cols where we expect to have a conflict while combining
        :param if_conflicts: The strategy to be applied if there are conflicts.
        """
        raise NotImplementedError

    def get_sqla_table(self, table: BaseTable) -> SqlaTable:
        """
        Return SQLAlchemy table instance

        :param table: Astro Table to be converted to SQLAlchemy table instance
        """
        return SqlaTable(
            table.name,
            table.sqlalchemy_metadata,
            autoload_with=self.sqlalchemy_engine,
            extend_existing=True,
        )

    # ---------------------------------------------------------
    # Extract methods
    # ---------------------------------------------------------
    def export_table_to_pandas_dataframe(
        self, source_table: BaseTable, select_kwargs: dict | None = None
    ) -> pd.DataFrame:
        """
        Copy the content of a table to an in-memory Pandas dataframe.

        :param source_table: An existing table in the database
        :param select_kwargs: kwargs for select statement
        """
        select_kwargs = select_kwargs or {}

        if self.table_exists(source_table):
            sqla_table = self.get_sqla_table(source_table)
            df = pd.read_sql(sql=sqla_table.select(**select_kwargs), con=self.sqlalchemy_engine)
            return PandasDataframe.from_pandas_df(df)

        table_qualified_name = self.get_table_qualified_name(source_table)
        raise NonExistentTableException(f"The table {table_qualified_name} does not exist")

    def export_table_to_file(
        self,
        source_table: BaseTable,
        target_file: File,
        if_exists: ExportExistsStrategy = "exception",
    ) -> None:
        """
        Copy the content of a table to a target file of supported type, in a supported location.

        :param source_table: An existing table in the database
        :param target_file: The path to the file to which we aim to dump the content of the database
        :param if_exists: Overwrite file if exists. Default False
        """
        if if_exists == "exception" and target_file.exists():
            raise FileExistsError(f"The file {target_file} already exists.")

        df = self.export_table_to_pandas_dataframe(source_table)
        target_file.create_from_dataframe(df)

    # ---------------------------------------------------------
    # Schema Management
    # ---------------------------------------------------------

    def create_schema_if_needed(self, schema: str | None) -> None:
        """
        This function checks if the expected schema exists in the database. If the schema does not exist,
        it will attempt to create it.

        :param schema: DB Schema - a namespace that contains named objects like (tables, functions, etc)
        """
        # We check if the schema exists first because snowflake will fail on a create schema query even if it
        # doesn't actually create a schema.
        if schema and not self.schema_exists(schema):
            statement = self._create_schema_statement.format(schema)
            self.run_sql(statement)

    def schema_exists(self, schema: str) -> bool:
        """
        Checks if a schema exists in the database

        :param schema: DB Schema - a namespace that contains named objects like (tables, functions, etc)
        """
        raise NotImplementedError

    # ---------------------------------------------------------
    # Context & Template Rendering methods (Transformations)
    # ---------------------------------------------------------

    def get_sqlalchemy_template_table_identifier_and_parameter(
        self, table: BaseTable, jinja_table_identifier: str  # skipcq PYL-W0613
    ) -> tuple[str, str]:
        """
        During the conversion from a Jinja-templated SQL query to a SQLAlchemy query, there is the need to
        convert a Jinja table identifier to a safe SQLAlchemy-compatible table identifier.

        For example, the query:
            sql = "SELECT * FROM {{input_table}};"
            parameters = {"input_table": Table(name="user_defined_table", metadata=Metadata(schema="some_schema"))}

        Can become (depending on the database):
            "SELECT * FROM some_schema.user_defined_table;"
            parameters = {"input_table": "user_defined_table"}

        Since the table value is templated, there is a safety concern (e.g. SQL injection).
        We recommend looking into the documentation of the database and seeing what are the best practices.
        For example, Snowflake:
        https://docs.snowflake.com/en/sql-reference/identifier-literal.html

        :param table: The table object we want to generate a safe table identifier for
        :param jinja_table_identifier: The name used within the Jinja template to represent this table
        :return: value to replace the table identifier in the query and the value that should be used to replace it
        """
        return (
            self.get_table_qualified_name(table),
            self.get_table_qualified_name(table),
        )

    def row_count(self, table: BaseTable):
        """
        Returns the number of rows in a table.

        :param table: table to count
        :return: The number of rows in the table
        """
        result = self.run_sql(
            f"select count(*) from {self.get_table_qualified_name(table)}"  # skipcq: BAN-B608
        ).scalar()
        return result

    def parameterize_variable(self, variable: str):
        """
        While most databases use sqlalchemy, we want to open up how we paramaterize variables for databases
        that a) do not use sqlalchemy and b) have different parameterization schemes (namely delta).

        :param variable: The variable to parameterize.
        :return: Variable with proper parameters
        """
        return ":" + variable

    def is_native_load_file_available(  # skipcq: PYL-R0201
        self, source_file: File, target_table: BaseTable  # skipcq: PYL-W0613
    ) -> bool:
        """
        Check if there is an optimised path for source to destination.

        :param source_file: File from which we need to transfer data
        :param target_table: Table that needs to be populated with file data
        """
        return False

    def load_file_to_table_natively(
        self,
        source_file: File,
        target_table: BaseTable,
        if_exists: LoadExistStrategy = "replace",
        native_support_kwargs: dict | None = None,
        **kwargs,
    ):
        """
        Checks if optimised path for transfer between File location to database exists
        and if it does, it transfers it and returns true else false

        :param source_file: File from which we need to transfer data
        :param target_table: Table that needs to be populated with file data
        :param if_exists: Overwrite file if exists. Default False
        :param native_support_kwargs: kwargs to be used by native loading command
        """
        raise NotImplementedError

    def check_schema_autodetection_is_supported(self, source_file: File) -> bool:
        """
        Checks if schema autodetection is handled natively by the database

        :param source_file: File from which we need to transfer data
        """
        filetype_supported = self.NATIVE_AUTODETECT_SCHEMA_CONFIG.get(source_file.location.location_type)

        source_filetype = (
            source_file
            if isinstance(source_file.type, FileTypeConstants)
            else create_file_type(path=source_file.path, filetype=source_file.type)  # type: ignore
        )

        is_source_filetype_supported = (
            (source_filetype.type.name in filetype_supported.get("filetype"))  # type: ignore
            if filetype_supported
            else None
        )

        location_type = self.NATIVE_PATHS.get(source_file.location.location_type)
        return bool(location_type and is_source_filetype_supported)

    def check_file_pattern_based_schema_autodetection_is_supported(self, source_file: File) -> bool:
        """
        Checks if schema autodetection is handled natively by the database for file
        patterns and prefixes.

        :param source_file: File from which we need to transfer data
        """
        is_file_pattern_based_schema_autodetection_supported = (
            source_file.location.location_type in self.FILE_PATTERN_BASED_AUTODETECT_SCHEMA_SUPPORTED
        )
        return is_file_pattern_based_schema_autodetection_supported

    def openlineage_dataset_name(self, table: BaseTable) -> str:
        """
        Returns the open lineage dataset namespace as per
        https://github.com/OpenLineage/OpenLineage/blob/main/spec/Naming.md
        """
        raise NotImplementedError

    def openlineage_dataset_namespace(self) -> str:
        """
        Returns the open lineage dataset namespace as per
        https://github.com/OpenLineage/OpenLineage/blob/main/spec/Naming.md
        """
        raise NotImplementedError

    def openlineage_dataset_uri(self, table: BaseTable) -> str:
        """
        Returns the open lineage dataset uri as per
        https://github.com/OpenLineage/OpenLineage/blob/main/spec/Naming.md
        """
        raise NotImplementedError



from __future__ import annotations

from typing import Any

from airflow.decorators.base import get_unique_task_id
from airflow.models.xcom_arg import XComArg

from astro.databases import create_database
from astro.sql.operators.base_operator import AstroSQLBaseOperator
from astro.table import BaseTable
from astro.utils.compat.typing import Context


class DropTableOperator(AstroSQLBaseOperator):
    """Airflow Operator for dropping SQL tables."""

    template_fields = ("table",)

    def __init__(
        self,
        table: BaseTable,
        task_id: str = "",
        **kwargs,
    ):
        self.table = table
        task_id = task_id or get_unique_task_id("drop")
        super().__init__(
            task_id=task_id,
            **kwargs,
        )

    def execute(self, context: Context) -> BaseTable:  # skipcq: PYL-W0613
        """Method run when the Airflow runner calls the operator."""
        database = create_database(self.table.conn_id, table=self.table)
        self.table = database.populate_table_metadata(self.table)
        database.drop_table(self.table)
        return self.table


def drop_table(
    table: BaseTable,
    **kwargs: Any,
) -> XComArg:
    """
    Drops a table.

    :param table: Table to be dropped
    :param kwargs: Any keyword arguments supported by the BaseOperator is supported (e.g ``queue``, ``owner``)
    """

    return DropTableOperator(table=table, **kwargs).output



from __future__ import annotations

import random
import string
from typing import TYPE_CHECKING

import pandas as pd

from astro import settings
from astro.constants import ColumnCapitalization, FileType
from astro.dataframes.pandas import PandasDataframe

if TYPE_CHECKING:
    from astro.files import File


def convert_columns_names_capitalization(
    df: pd.DataFrame, columns_names_capitalization: ColumnCapitalization
):
    """
    Convert cols of a dataframe to required case. Options - lower/Upper

    :param df: dataframe whose cols will be altered
    :param columns_names_capitalization: String Literal with possible values - lower/Upper
    """
    if isinstance(df, pd.DataFrame):
        df = PandasDataframe.from_pandas_df(df)
        if columns_names_capitalization == "lower":
            df.columns = [col_label.lower() for col_label in df.columns]
        elif columns_names_capitalization == "upper":
            df.columns = [col_label.upper() for col_label in df.columns]

    return df


def convert_dataframe_to_file(df: pd.DataFrame) -> File:
    """
    Passes a dataframe into a File using parquet as an efficient storage format. This allows us to use
    Json as a storage method without filling the metadata database. the values for conn_id and bucket path can
    be found in the airflow.cfg as follows:

    [astro]
    xcom_storage_conn_id=...
    xcom_storage_url=///
    :param df: Dataframe to convert to file
    :return: File object with reference to stored dataframe file
    """
    unique_id = random.choice(string.ascii_lowercase) + "".join(
        random.choice(string.ascii_lowercase + string.digits) for _ in range(64)
    )

    # importing here to prevent circular imports
    from astro.files import File

    file = File(
        path=settings.DATAFRAME_STORAGE_URL + "/" + unique_id + ".parquet",
        conn_id=settings.DATAFRAME_STORAGE_CONN_ID,
        filetype=FileType.PARQUET,
        is_dataframe=True,
    )
    file.create_from_dataframe(df)
    return file



from __future__ import annotations

from urllib.parse import urlparse
from urllib.request import urlopen

from astro.constants import FileLocation
from astro.files.locations.base import BaseFileLocation


class HTTPLocation(BaseFileLocation):
    """Handler http location operations"""

    location_type = FileLocation.HTTP

    @property
    def paths(self) -> list[str]:
        """Resolve patterns in path"""
        return [self.path]

    def validate_conn(self):
        """Override as conn_id is not always required for http location."""

    @property
    def size(self) -> int:
        """Return file size for HTTP location"""
        file = urlopen(self.path)  # skipcq BAN-B310
        return int(file.length)

    @property
    def openlineage_dataset_namespace(self) -> str:
        """
        Returns the open lineage dataset namespace as per
        https://github.com/OpenLineage/OpenLineage/blob/main/spec/Naming.md
        """
        parsed_url = urlparse(self.path)
        return f"{parsed_url.scheme}://{parsed_url.netloc}"

    @property
    def openlineage_dataset_name(self) -> str:
        """
        Returns the open lineage dataset name as per
        https://github.com/OpenLineage/OpenLineage/blob/main/spec/Naming.md
        """
        return urlparse(self.path).path



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING

import pandas as pd
import sqlalchemy
from airflow.providers.microsoft.mssql.hooks.mssql import MsSqlHook
from pymssql._pymssql import ProgrammingError
from sqlalchemy.sql import ClauseElement

from astro.constants import DEFAULT_CHUNK_SIZE, LoadExistStrategy, MergeConflictStrategy
from astro.databases.base import BaseDatabase
from astro.options import LoadOptions
from astro.settings import MSSQL_SCHEMA
from astro.table import BaseTable, Metadata
from astro.utils.compat.functools import cached_property

DEFAULT_CONN_ID = MsSqlHook.default_conn_name
if TYPE_CHECKING:  # pragma: no cover
    from sqlalchemy.engine.cursor import CursorResult


class MssqlDatabase(BaseDatabase):
    DEFAULT_SCHEMA = MSSQL_SCHEMA

    _create_schema_statement: str = """
    IF NOT EXISTS (SELECT 1 FROM sys.schemas WHERE name = '{0}')
    BEGIN
        EXEC( 'CREATE SCHEMA {0}' );
    END
    """
    _truncate_table_statement: str = "TRUNCATE TABLE {0}"

    # CTAS is not supported in sql server. Hence, use select * into
    # select * into from a subquery will work in sql server only when an alias is added
    _create_table_statement: str = """
    IF (NOT EXISTS (SELECT *
                 FROM INFORMATION_SCHEMA.TABLES
                 WHERE TABLE_SCHEMA = '{0}'
                 AND   TABLE_NAME    = '{1}'
                 AND   TABLE_CATALOG = '{2}'))
    BEGIN
        SELECT * INTO  {2}.{0}.{1} FROM ({3}) as astro;
    END
    """

    # alter , select * into, drop table are considered as DDL(they change the system catalog).
    # sqlalchemy based on mssql dialect considers autocommit as False by default which causes these
    # statements to not execute

    _queries_requiring_autocommit: list = ["alter", "select * into"]

    def __init__(
        self,
        conn_id: str = DEFAULT_CONN_ID,
        table: BaseTable | None = None,
        load_options: LoadOptions | None = None,
    ):
        super().__init__(conn_id)
        self.table = table
        self.load_options = load_options

    @cached_property
    def hook(self) -> MsSqlHook:
        """Retrieve Airflow hook to interface with the mssql database."""
        kwargs = {}
        _hook = MsSqlHook(mssql_conn_id=self.conn_id)
        if (_hook.schema is None) and (self.table and self.table.metadata and self.table.metadata.database):
            kwargs.update({"schema": self.table.metadata.database})
        return MsSqlHook(mssql_conn_id=self.conn_id, **kwargs)

    @property
    def sql_type(self) -> str:
        return "mssql"

    def table_exists(self, table: BaseTable) -> bool:
        """
        Check if a table exists in the database

        :param table: Details of the table we want to check that exists
        """
        inspector = sqlalchemy.inspect(self.sqlalchemy_engine)

        schema = None
        if table and table.metadata:
            schema = table.metadata.schema
        return bool(inspector.dialect.has_table(self.connection, table.name, schema=schema))

    @staticmethod
    def get_table_qualified_name(table: BaseTable) -> str:
        """
        Return the table qualified name.

        :param table: The table we want to retrieve the qualified name for.
        """
        qualified_name_lists = [
            table.metadata.database,
            table.metadata.schema,
            table.name,
        ]
        qualified_name = ".".join(name for name in qualified_name_lists if name)
        return qualified_name

    @property
    def default_metadata(self) -> Metadata:
        """Fill in default metadata values for table objects addressing mssql databases.
        Currently, Schema is not being fetched from airflow connection for Mssql because, in Mssql,
        databases and schema are different concepts
        The MssqlHook only exposes schema
        However, implementation-wise, it seems that if the MssqlHook receives a schema during
        initialization, but it uses it as a database in the connection to Mssql:
        """
        # TODO: Change airflow MssqlHook to fetch database and schema separately
        database = self.hook.get_connection(self.conn_id).schema
        return Metadata(database=database, schema=self.DEFAULT_SCHEMA)  # type: ignore

    def schema_exists(self, schema) -> bool:
        """
        Checks if a schema exists in the database

        :param schema: DB Schema - a namespace that contains named objects like (tables, functions, etc)
        """
        try:
            schema_result = self.hook.run(
                "SELECT schema_name FROM information_schema.schemata WHERE "
                "lower(schema_name) = lower(%(schema_name)s);",
                parameters={"schema_name": schema.lower()},
                handler=lambda x: [y[0] for y in x.fetchall()],
            )
            return len(schema_result) > 0
        except ProgrammingError:
            return False

    def is_autocommit_required(self, sql) -> bool:
        """
        Checks if autocommit is required for a query

        :param sql: sql query which needs to be checked for autocommit setting

        """
        return any(query in sql.lower() for query in self._queries_requiring_autocommit)

    def run_sql(
        self,
        sql: str | ClauseElement = "",
        parameters: dict | None = None,
        **kwargs,
    ) -> CursorResult:
        """
        Return the results to running a SQL statement.
        Whenever possible, this method should be implemented using Airflow Hooks,
        since this will simplify the integration with Async operators.

        :param sql: Contains SQL query to be run against database
        :param parameters: Optional parameters to be used to render the query
        """
        if parameters is None:
            parameters = {}

        if "sql_statement" in kwargs:  # pragma: no cover
            warnings.warn(
                "`sql_statement` is deprecated and will be removed in future release"
                "Please use  `sql` param instead.",
                DeprecationWarning,
                stacklevel=2,
            )
            sql = kwargs.get("sql_statement")  # type: ignore

        autocommit = kwargs.get("autocommit", False)

        # We need to set autocommit=True for specific queries
        if isinstance(sql, str):
            if (autocommit is True) or self.is_autocommit_required(sql):
                self.connection.execute(sqlalchemy.text(sql).execution_options(autocommit=True), parameters)
            else:
                result = self.connection.execute(
                    sqlalchemy.text(sql).execution_options(autocommit=autocommit), parameters
                )
                return result
        else:
            # this is used for append
            result = self.connection.execute(sql, parameters)
            return result

    def create_schema_if_needed(self, schema: str | None) -> None:
        """
        This function checks if the expected schema exists in the database. If the schema does not exist,
        it will attempt to create it.

        :param schema: DB Schema - a namespace that contains named objects like (tables, functions, etc)
        """
        if schema and not self.schema_exists(schema):
            statement = self._create_schema_statement.format(schema)
            self.run_sql(statement, autocommit=True)

    def create_table_from_select_statement(
        self,
        statement: str,
        target_table: BaseTable,
        parameters: dict | None = None,
    ) -> None:
        """
        Export the result rows of a query statement into another table.

        :param statement: SQL query statement
        :param target_table: Destination table where results will be recorded.
        :param parameters: (Optional) parameters to be used to render the SQL query
        """

        statement = self._create_table_statement.format(
            target_table.metadata.schema,
            target_table.name,
            target_table.metadata.database,
            statement,
        )

        self.run_sql(statement, parameters, autocommit=True)

    def drop_table(self, table: BaseTable) -> None:
        """
        Delete a SQL table, if it exists.

        :param table: The table to be deleted.
        """
        statement = self._drop_table_statement.format(self.get_table_qualified_name(table))
        self.run_sql(statement, autocommit=True)

    def fetch_all_rows(self, table: BaseTable, row_limit: int = -1) -> list:
        """
        Fetches all rows for a table and returns as a list. This is needed because some
        databases have different cursors that require different methods to fetch rows

        :param row_limit: Limit the number of rows returned, by default return all rows.
        :param table: The table metadata needed to fetch the rows
        :return: a list of rows
        """
        statement = f"SELECT * FROM {self.get_table_qualified_name(table)}"  # skipcq: BAN-B608
        if row_limit > -1:
            statement = f"SELECT TOP {row_limit} * FROM {self.get_table_qualified_name(table)}"
        response = self.run_sql(statement)
        return response.fetchall()  # type: ignore

    def load_pandas_dataframe_to_table(
        self,
        source_dataframe: pd.DataFrame,
        target_table: BaseTable,
        if_exists: LoadExistStrategy = "replace",
        chunk_size: int = DEFAULT_CHUNK_SIZE,
    ) -> None:  # skipcq PYL-W0613
        """
        Create a table with the dataframe's contents.
        If the table already exists, append or replace the content, depending on the value of `if_exists`.

        :param source_dataframe: Local or remote filepath
        :param target_table: Table in which the file will be loaded
        :param if_exists: Strategy to be used in case the target table already exists.
        :param chunk_size: Specify the number of rows in each batch to be written at a time.
        """
        self._assert_not_empty_df(source_dataframe)

        self.create_schema_if_needed(target_table.metadata.schema)
        if not self.table_exists(table=target_table) or if_exists == "replace":
            self.create_table(table=target_table, dataframe=source_dataframe)

        table_name = self.get_table_qualified_name(target_table)
        mssql_conn = self.hook.get_conn()
        source_dataframe = source_dataframe.astype(object).where(pd.notnull(source_dataframe), None)
        mssql_conn.bulk_copy(table_name, list(source_dataframe.itertuples(index=False, name=None)))
        mssql_conn.commit()

    def merge_table(
        self,
        source_table: BaseTable,
        target_table: BaseTable,
        source_to_target_columns_map: dict[str, str],
        target_conflict_columns: list[str],
        if_conflicts: MergeConflictStrategy = "exception",
    ) -> None:
        """
        Merge the source table rows into a destination table.
        The argument `if_conflicts` allows the user to define how to handle conflicts.

        :param source_table: Contains the rows to be merged to the target_table
        :param target_table: Contains the destination table in which the rows will be merged
        :param source_to_target_columns_map: Dict of target_table columns names to source_table columns names
        :param target_conflict_columns: List of cols where we expect to have a conflict while combining
        :param if_conflicts: The strategy to be applied if there are conflicts.
        """

        statement, params = self._build_merge_sql(
            source_table=source_table,
            target_table=target_table,
            source_to_target_columns_map=source_to_target_columns_map,
            target_conflict_columns=target_conflict_columns,
            if_conflicts=if_conflicts,
        )
        # sql server generates nvarchar strings when passing parameter values to the query
        # Hence replace the params here instead of letting sql server doing it

        for k, v in params.items():
            statement = statement.replace(k, v)
        self.run_sql(sql=statement, parameters={}, autocommit=True)

    @staticmethod
    def get_merge_initialization_query(parameters: tuple) -> str:
        """
        Handles database-specific logic to handle constraints, keeping
        it agnostic to database.
        """
        sql = "RETURN"
        return sql

    def _build_merge_sql(
        self,
        source_table: BaseTable,
        target_table: BaseTable,
        source_to_target_columns_map: dict[str, str],
        target_conflict_columns: list[str],
        if_conflicts: MergeConflictStrategy = "exception",
    ):
        """Build the SQL statement for Merge operation"""
        source_table_name = source_table.name
        target_table_name = target_table.name

        source_cols = source_to_target_columns_map.keys()
        target_cols = source_to_target_columns_map.values()

        target_identifier_enclosure = ""
        source_identifier_enclosure = ""

        (
            source_table_identifier,
            source_table_param,
        ) = self.get_sqlalchemy_template_table_identifier_and_parameter(source_table, "source_table")

        (
            target_table_identifier,
            target_table_param,
        ) = self.get_sqlalchemy_template_table_identifier_and_parameter(target_table, "target_table")

        statement = (
            f"merge {target_table_identifier} as target using {source_table_identifier} as source "
            "on {merge_clauses}"
        )

        merge_target_dict = {
            f"merge_clause_target_{i}": f"target.{x}" for i, x in enumerate(target_conflict_columns)
        }
        merge_source_dict = {
            f"merge_clause_source_{i}": f"source.{x}" for i, x in enumerate(target_conflict_columns)
        }

        statement = statement.replace(
            "{merge_clauses}",
            " AND ".join(
                f"{wrap_identifier(k)}={wrap_identifier(v)}"
                for k, v in zip(merge_target_dict.keys(), merge_source_dict.keys())
            ),
        )

        values_to_check = [target_table_name, source_table_name]
        values_to_check.extend(source_cols)
        values_to_check.extend(target_cols)
        if if_conflicts == "update":
            statement += " when matched then UPDATE SET {merge_vals}"
            merge_statement = ",".join(
                [
                    f"target.{target_identifier_enclosure}{t}{target_identifier_enclosure}="
                    f"source.{source_identifier_enclosure}{s}{source_identifier_enclosure}"
                    for s, t in source_to_target_columns_map.items()
                ]
            )
            statement = statement.replace("{merge_vals}", merge_statement)
        statement += " when not matched by target then insert({target_columns}) values ({append_columns}); "
        statement = statement.replace(
            "{target_columns}",
            ",".join(f"{target_identifier_enclosure}{t}{target_identifier_enclosure}" for t in target_cols),
        )
        statement = statement.replace(
            "{append_columns}",
            ",".join(
                f"source.{source_identifier_enclosure}{s}{source_identifier_enclosure}" for s in source_cols
            ),
        )

        params = {
            **merge_target_dict,
            **merge_source_dict,
            "source_table": source_table_param,
            "target_table": target_table_param,
        }
        return statement, params


def wrap_identifier(inp: str) -> str:
    return f"{inp}"



from __future__ import annotations

import io

import pandas as pd

from astro.constants import FileType as FileTypeConstants
from astro.dataframes.load_options import PandasLoadOptions
from astro.dataframes.pandas import PandasDataframe
from astro.files.types.base import FileType
from astro.utils.dataframe import convert_columns_names_capitalization


class JSONFileType(FileType):
    """Concrete implementation to handle JSON file type"""

    LOAD_OPTIONS_CLASS_NAME = "PandasJsonLoadOptions"

    # We need skipcq because it's a method overloading so we don't want to make it a static method
    def export_to_dataframe(
        self,
        stream: io.TextIOWrapper,
        columns_names_capitalization="original",
        **kwargs,
    ) -> pd.DataFrame:  # skipcq PYL-R0201
        """read json file from one of the supported locations and return dataframe

        :param stream: file stream object
        :param load_options: Pandas option to pass to the Pandas lib while reading json
        :param columns_names_capitalization: determines whether to convert all columns to lowercase/uppercase
            in the resulting dataframe
        """
        kwargs_copy = dict(kwargs)
        # Pandas `read_json` does not support the `nrows` parameter unless we're using NDJSON
        kwargs_copy.pop("nrows", None)
        if isinstance(self.load_options, PandasLoadOptions):
            kwargs_copy.update(self.load_options.to_dict())
        df = pd.read_json(stream, **kwargs_copy)
        df = convert_columns_names_capitalization(
            df=df, columns_names_capitalization=columns_names_capitalization
        )
        return PandasDataframe.from_pandas_df(df)

    # We need skipcq because it's a method overloading so we don't want to make it a static method
    def create_from_dataframe(self, df: pd.DataFrame, stream: io.TextIOWrapper) -> None:  # skipcq PYL-R0201
        """Write json file to one of the supported locations

        :param df: pandas dataframe
        :param stream: file stream object
        """
        df.to_json(stream, orient="records")

    @property
    def name(self):
        return FileTypeConstants.JSON



from __future__ import annotations

from attr import define
from pandas._typing import DtypeArg

from astro.options import LoadOptions


@define
class PandasLoadOptions(LoadOptions):
    """Pandas load options while reading and loading file"""

    pass


@define
class PandasCsvLoadOptions(PandasLoadOptions):
    """
    Pandas load options while reading and loading csv file.

    :param delimiter: Delimiter to use. Defaults to None
    :param dtype: Data type for data or columns. E.g. ``{"a": np.float64, "b": np.int32, "c": "Int64"}`` Use str or
        object together with suitable na_values settings to preserve and not interpret dtype. If converters are
        specified, they will be applied INSTEAD of dtype conversion.
    """

    delimiter: str | None = None
    dtype: DtypeArg | None = None


@define
class PandasJsonLoadOptions(PandasLoadOptions):
    """
    Pandas load options while reading and loading json file.

    :param encoding: Encoding to use for UTF when reading/writing (ex. ‘utf-8’).
        List of Python standard encodings: https://docs.python.org/3/library/codecs.html#standard-encodings
    """

    encoding: str | None = None


@define
class PandasNdjsonLoadOptions(PandasLoadOptions):
    """
    Pandas load options while reading and loading Ndjson file.

    :param normalize_sep: separator used to normalize nested ndjson.
         ex - ``{"a": {"b":"c"}}`` will result in: ``column - "a_b"`` where ``ndjson_normalize_sep = "_"``
    """

    normalize_sep: str = "_"


@define
class PandasParquetLoadOptions(PandasLoadOptions):
    """
    Pandas load options while reading and loading Parquet file.

    :param columns: If not None, only these columns will be read from the file.
    """

    columns: list[str] | None = None
